from __future__ import division

import copy
import logging
import os
import pdb
from collections import Counter, defaultdict
from timeit import default_timer as timer

import pandas as pd
import pyprind
import six
from sklearn.decomposition import TruncatedSVD

import torch
import torch.nn as nn
from torchtext.legacy import data
from torchtext.legacy.data.example import Example
from torchtext.utils import unicode_csv_reader

from ..models.modules import NoMeta, Pool
from .field import MatchingField
from .iterator import MatchingIterator

logger = logging.getLogger(__name__)


def split(table,
          path,
          train_prefix,
          validation_prefix,
          test_prefix,
          split_ratio=[0.6, 0.2, 0.2],
          stratified=False,
          strata_field='label'):
    """Split a pandas dataframe or CSV file into train / validation / test data sets.

    Args:
        table (pandas.Dataframe or string): The pandas dataframe or CSV file to split.
        path (string): The directory to save the train, validation and test CSV files to.
        train: Suffix to add to `path` to get the training set save path.
        validation: Suffix to add to `path` to get the validation set save path.
        test: Suffix to add to `path` to get the test set save path.
        split_ratio (List of floats): a list of 3 numbers denoting the relative sizes of
            train, test and valid splits respectively. Default is [0.6, 0.2, 0.2].
        stratified (bool): whether the sampling should be stratified.
            Default is False.
        strata_field (str): name of the examples Field stratified over.
            Default is 'label' for the conventional label field.
    """
    assert len(split_ratio) == 3

    if not isinstance(table, pd.DataFrame):
        table = pd.read_csv(table)
    if table.index.name is not None:
        table = table.reset_index()

    examples = list(table.itertuples(index=False))
    fields = [(col, None) for col in list(table)]
    dataset = data.Dataset(examples, fields)
    train, valid, test = dataset.split(split_ratio, stratified, strata_field)

    tables = (pd.DataFrame(train.examples), pd.DataFrame(valid.examples),
              pd.DataFrame(test.examples))
    prefixes = (train_prefix, validation_prefix, test_prefix)

    for i in range(len(tables)):
        tables[i].columns = table.columns
        tables[i].to_csv(os.path.join(path, prefixes[i]), index=False)


class MatchingDataset(data.Dataset):
    r"""Represents dataset with associated metadata.

    Holds all information about one split of a dataset (e.g. training set).

    Attributes:
        fields (dict): A mapping from attribute names (e.g. "left_address") to
            corresponding :class:`~data.MatchingField` objects that specify how to process
            the field.
        examples (list): A list containing all the examples (labeled tuple pairs) in this
            dataset.
        metadata (dict): Metadata about the dataset (e.g. word probabilities).
            See :meth:`~data.MatchingDataset.compute_metadata` for details.
        corresponding_field (dict): A mapping from left table attribute names
            (e.g. "left_address") to corresponding right table attribute names
            (e.g. "right_address") and vice versa.
        text_fields (dict): A mapping from canonical attribute names (e.g. "address") to
            tuples of the corresponding left and right attribute names
            (e.g. ("left_address", "right_address")).
        all_left_fields (list): A list of all left table attribute names.
        all_right_fields (list): A list of all right table attribute names.
        canonical_text_fields (list): A list of all canonical attribute names.
        label_field (str): Name of the column containing labels.
        id_field (str): Name of the column containing tuple pair ids.
    """

    class CacheStaleException(Exception):
        r"""Raised when the dataset cache is stale and no fallback behavior is specified.
        """
        pass

    def __init__(self,
                 fields,
                 column_naming,
                 path=None,
                 format='csv',
                 examples=None,
                 metadata=None,
                 **kwargs):
        r"""Creates a MatchingDataset.

        Creates a MatchingDataset by performing the following, if `examples` parameter is
        not specified:

        #. Read each example (tuple pair) in specified CSV file.
        #. Preprocess example. Involves lowercasing and tokenization (unless disabled).
        #. Compute metadata. See :meth:`~data.MatchingDataset.compute_metadata` for
            details.

        If `examples` is specified, initializes MatchingDataset from given `examples`
        and `metadata` arguments.

        Arguments:
            fields (list(tuple(str, MatchingField))): A list of tuples containing column
                name (e.g. "left_address") and corresponding :class:`~data.MatchingField`
                pairs, in the same order that the columns occur in the CSV file. Tuples of
                (name, None) represent columns that will be ignored.
            column_naming (dict): A `dict` containing the following keys:
                * ``id``: The name of the tuple pair ID column.
                * ``label``: The name of the tuple pair match label column.
                * ``left``: The prefix for attribute names belonging to the left table.
                * ``right``: The prefix for attribute names belonging to the right table.
            path (str): Path to the data file. Must be specified if `examples` is None.
            format (str): The format of the data file. One of "CSV" or "TSV".
            examples (list): A list containing all the examples (labeled tuple pairs) in
                this dataset. Must be specified if `path` is None.
            metadata (dict): Metadata about the dataset (e.g. word probabilities).
                See :meth:`~data.MatchingDataset.compute_metadata` for details.
            filter_pred (callable or None): Use only examples for which
                filter_pred(example) is True, or use all examples if None.
                Default is None. This is a keyword-only parameter.
        """
        if examples is None:
            make_example = {
                'json': Example.fromJSON, 'dict': Example.fromdict,
                'tsv': Example.fromCSV, 'csv': Example.fromCSV}[format.lower()]

            lines = 0
            with open(os.path.expanduser(path), encoding="utf8") as f:
                for line in f:
                    lines += 1

            with open(os.path.expanduser(path), encoding="utf8") as f:
                if format == 'csv':
                    reader = unicode_csv_reader(f)
                elif format == 'tsv':
                    reader = unicode_csv_reader(f, delimiter='\t')
                else:
                    reader = f

                next(reader)
                examples = [make_example(line, fields) for line in
                    pyprind.prog_bar(reader, iterations=lines,
                        title='\nReading and processing data from "' + path + '"')]

            super(MatchingDataset, self).__init__(examples, fields, **kwargs)
        else:
            self.fields = dict(fields)
            self.examples = examples
            self.metadata = metadata

        self.path = path
        self.column_naming = column_naming
        self._set_attributes()

    def _set_attributes(self):
        """Sets attributes by inferring mapping between left and right table attributes.
        """
        self.corresponding_field = {}
        self.text_fields = {}

        self.all_left_fields = []
        for name, field in six.iteritems(self.fields):
            if name.startswith(self.column_naming['left']) and field is not None:
                self.all_left_fields.append(name)

        self.all_right_fields = []
        for name, field in six.iteritems(self.fields):
            if name.startswith(self.column_naming['right']) and field is not None:
                self.all_right_fields.append(name)

        self.canonical_text_fields = []
        for left_name in self.all_left_fields:
            canonical_name = left_name[len(self.column_naming['left']):]
            right_name = self.column_naming['right'] + canonical_name
            self.corresponding_field[left_name] = right_name
            self.corresponding_field[right_name] = left_name
            self.text_fields[canonical_name] = left_name, right_name
            self.canonical_text_fields.append(canonical_name)

        self.all_text_fields = self.all_left_fields + self.all_right_fields
        self.label_field = self.column_naming['label']
        self.id_field = self.column_naming['id']

    def compute_metadata(self, pca=False):
        r"""Computes metadata about the dataset.

        Computes the following metadata about the dataset:

        * ``word_probs``: For each attribute in the dataset, a mapping from words to
          word (token) probabilities.
        * ``totals``: For each attribute in the dataset, a count of the total number of
          words present in all attribute examples.
        * ``pc``: For each attribute in the dataset, the first principal component of the
          sequence embeddings for all values of that attribute. The sequence embedding of
          an attribute value is computed by taking the weighted average of its word
          embeddings, where the weight is the soft inverse word probability. Refer
          `Arora et al. (2017) <https://openreview.net/pdf?id=SyK00v5xx>`__ for details.

        Arguments:
            pca (bool): Whether to compute the ``pc`` metadata.
        """
        self.metadata = {}

        # Create an iterator over the entire dataset.
        train_iter = MatchingIterator(
            self, self, train=False, batch_size=1024, device='cpu', sort_in_buckets=False)
        counter = defaultdict(Counter)

        # For each attribute, find the number of times each word id occurs in the dataset.
        # Note that word ids here also include ``UNK`` tokens, padding tokens, etc.
        for batch in pyprind.prog_bar(train_iter, title='\nBuilding vocabulary'):
            for name in self.all_text_fields:
                attr_input = getattr(batch, name)
                counter[name].update(attr_input.data.data.view(-1).tolist())

        word_probs = {}
        totals = {}
        for name in self.all_text_fields:
            attr_counter = counter[name]
            total = sum(attr_counter.values())
            totals[name] = total

            field_word_probs = {}
            for word, freq in attr_counter.items():
                field_word_probs[word] = freq / total
            word_probs[name] = field_word_probs
        self.metadata['word_probs'] = word_probs
        self.metadata['totals'] = totals

        if not pca:
            return

        # To compute principal components, we need to compute weighted sequence embeddings
        # for each attribute. To do so, for each attribute, we first construct a neural
        # network to compute word embeddings and take their weighted average.
        field_embed = {}
        embed = {}
        inv_freq_pool = Pool('inv-freq-avg')
        for name in self.all_text_fields:
            field = self.fields[name]
            if field not in field_embed:
                vectors_size = field.vocab.vectors.shape
                embed_layer = nn.Embedding(vectors_size[0], vectors_size[1])
                embed_layer.weight.data.copy_(field.vocab.vectors)
                embed_layer.weight.requires_grad = False
                field_embed[field] = NoMeta(embed_layer)
            embed[name] = field_embed[field]

        # Create an iterator over the entire dataset.
        train_iter = MatchingIterator(
            self, self, train=False, batch_size=1024, device='cpu', sort_in_buckets=False)
        attr_embeddings = defaultdict(list)

        # Run the constructed neural network to compute weighted sequence embeddings
        # for each attribute of each example in the dataset.
        for batch in pyprind.prog_bar(train_iter,
            title='\nComputing principal components'):
            for name in self.all_text_fields:
                attr_input = getattr(batch, name)
                embeddings = inv_freq_pool(embed[name](attr_input))
                attr_embeddings[name].append(embeddings.data.data)

        # Compute the first principal component of weighted sequence embeddings for each
        # attribute.
        pc = {}
        for name in self.all_text_fields:
            concatenated = torch.cat(attr_embeddings[name])
            svd = TruncatedSVD(n_components=1, n_iter=7)
            svd.fit(concatenated.numpy())
            pc[name] = svd.components_[0]
        self.metadata['pc'] = pc

    def finalize_metadata(self):
        r"""Perform final touches to dataset metadata.

        This allows performing modifications to metadata that cannot be serialized into
        the cache.
        """

        self.orig_metadata = copy.deepcopy(self.metadata)
        for name in self.all_text_fields:
            self.metadata['word_probs'][name] = defaultdict(
                lambda: 1 / self.metadata['totals'][name],
                self.metadata['word_probs'][name])

    def get_raw_table(self):
        r"""Create a raw pandas table containing all examples (tuple pairs) in the dataset.

        To resurrect tokenized attributes, this method currently naively joins the tokens
        using the whitespace delimiter.
        """
        rows = []
        columns = list(name for name, field in six.iteritems(self.fields) if field)
        for ex in self.examples:
            row = []
            for attr in columns:
                if self.fields[attr]:
                    val = getattr(ex, attr)
                    if self.fields[attr].sequential:
                        val = ' '.join(val)
                    row.append(val)
            rows.append(row)

        return pd.DataFrame(rows, columns=columns)

    def sort_key(self, ex):
        r"""Sort key for dataset examples.

        A key to use for sorting dataset examples for batching together examples with
        similar lengths to minimize padding."""

        return interleave_keys([len(getattr(ex, attr)) for attr in self.all_text_fields])

    @staticmethod
    def save_cache(datasets, fields, datafiles, cachefile, column_naming, state_args):
        r"""Save datasets and corresponding metadata to cache.

        This method also saves as many data loading arguments as possible to help ensure
        that the cache contents are still relevant for future data loading calls. Refer
        to :meth:`~data.Dataset.load_cache` for more details.

        Arguments:
            datasets (list): List of datasets to cache.
            fields (dict): Mapping from attribute names (e.g. "left_address") to
                corresponding :class:`~data.MatchingField` objects that specify how to
                process the field.
            datafiles (list): A list of the data files.
            cachefile (str): The cache file path.
            column_naming (dict): A `dict` containing column naming conventions. See
                `__init__` for details.
            state_args (dict): A `dict` containing other information about the state under
                which the cache was created.
        """
        examples = [dataset.examples for dataset in datasets]
        train_metadata = datasets[0].metadata
        datafiles_modified = [os.path.getmtime(datafile) for datafile in datafiles]
        vocabs = {}
        field_args = {}
        reverse_fields = {}
        for name, field in six.iteritems(fields):
            reverse_fields[field] = name

        for field, name in six.iteritems(reverse_fields):
            if field is not None and hasattr(field, 'vocab'):
                vocabs[name] = field.vocab
        for name, field in six.iteritems(fields):
            field_args[name] = None
            if field is not None:
                field_args[name] = field.preprocess_args()

        data = {
            'examples': examples,
            'train_metadata': train_metadata,
            'vocabs': vocabs,
            'datafiles': datafiles,
            'datafiles_modified': datafiles_modified,
            'field_args': field_args,
            'state_args': state_args,
            'column_naming': column_naming
        }
        torch.save(data, cachefile)

    @staticmethod
    def load_cache(fields, datafiles, cachefile, column_naming, state_args):
        r"""Load datasets and corresponding metadata from cache.

        This method also checks whether any of the data loading arguments have changes
        that make the cache contents invalid. The following kinds of changes are currently
        detected automatically:

        * Data filename changes (e.g. different train filename)
        * Data file modifications (e.g. train data modified)
        * Column changes (e.g. using a different subset of columns in CSV file)
        * Column specification changes (e.g. changing lowercasing behavior)
        * Column naming convention changes (e.g. different labeled data column)

        Arguments:
            fields (dict): Mapping from attribute names (e.g. "left_address") to
                corresponding :class:`~data.MatchingField` objects that specify how to
                process the field.
            datafiles (list): A list of the data files.
            cachefile (str): The cache file path.
            column_naming (dict): A `dict` containing column naming conventions. See
                `__init__` for details.
            state_args (dict): A `dict` containing other information about the state under
                which the cache was created.

        Returns:
            Tuple containing unprocessed cache data dict and a list of cache staleness
            causes, if any.

        .. warning::
            Note that if a column specification, i.e., arguments to
            :class:`~data.MatchingField` include callable arguments (e.g. lambdas or
            functions) these arguments cannot be serialized and hence will not be checked
            for modifications.
        """
        cached_data = torch.load(cachefile)
        cache_stale_cause = set()

        if datafiles != cached_data['datafiles']:
            cache_stale_cause.add('Data file list has changed.')

        datafiles_modified = [os.path.getmtime(datafile) for datafile in datafiles]
        if datafiles_modified != cached_data['datafiles_modified']:
            cache_stale_cause.add('One or more data files have been modified.')

        if set(fields.keys()) != set(cached_data['field_args'].keys()):
            cache_stale_cause.add('Fields have changed.')

        for name, field in six.iteritems(fields):
            none_mismatch = (field is None) != (cached_data['field_args'][name] is None)
            args_mismatch = False
            if field is not None and cached_data['field_args'][name] is not None:
                args_mismatch = field.preprocess_args() != cached_data['field_args'][name]
            if none_mismatch or args_mismatch:
                cache_stale_cause.add('Field arguments have changed.')
            if field is not None and not isinstance(field, MatchingField):
                cache_stale_cause.add('Cache update required.')

        if column_naming != cached_data['column_naming']:
            cache_stale_cause.add('Other arguments have changed.')

        cache_stale_cause.update(
            MatchingDataset.state_args_compatibility(state_args,
                                                     cached_data['state_args']))

        return cached_data, cache_stale_cause

    @staticmethod
    def state_args_compatibility(cur_state, old_state):
        errors = []
        if not old_state['train_pca'] and cur_state['train_pca']:
            errors.append('PCA computation necessary.')
        return errors

    @staticmethod
    def restore_data(fields, cached_data):
        r"""Recreate datasets and related data from cache.

        This restores all datasets, metadata and attribute information (including the
        vocabulary and word embeddings for all tokens in each attribute).
        """
        datasets = []
        for d in range(len(cached_data['datafiles'])):
            metadata = None
            if d == 0:
                metadata = cached_data['train_metadata']
            dataset = MatchingDataset(
                path=cached_data['datafiles'][d],
                fields=fields,
                examples=cached_data['examples'][d],
                metadata=metadata,
                column_naming=cached_data['column_naming'])
            datasets.append(dataset)

        for name, field in fields:
            if name in cached_data['vocabs']:
                field.vocab = cached_data['vocabs'][name]

        return datasets

    @classmethod
    def splits(cls,
               path,
               train=None,
               validation=None,
               test=None,
               fields=None,
               embeddings=None,
               embeddings_cache=None,
               column_naming=None,
               cache=None,
               check_cached_data=True,
               auto_rebuild_cache=False,
               train_pca=False,
               **kwargs):
        r"""Create Dataset objects for multiple splits of a dataset.

        Args:
            path (str): Common prefix of the splits' file paths.
            train (str): Suffix to add to path for the train set.
            validation (str): Suffix to add to path for the validation set, or None
                for no validation set. Default is None.
            test (str): Suffix to add to path for the test set, or None for no test
                set. Default is None.
            fields (list(tuple(str, MatchingField))): A list of tuples containing column
                name (e.g. "left_address") and corresponding :class:`~data.MatchingField`
                pairs, in the same order that the columns occur in the CSV file. Tuples of
                (name, None) represent columns that will be ignored.
            embeddings (str or list): Same as `embeddings` parameter of
                :func:`~data.process`.
            embeddings_cache (str): Directory to store dowloaded word vector data.
            column_naming (dict): Same as `column_naming` paramter of `__init__`.
            cache (str): Suffix to add to path for cache file. If `None` disables caching.
            check_cached_data (bool): Verify that data files haven't changes since the
                cache was constructed and that relevant field options haven't changed.
            auto_rebuild_cache (bool): Automatically rebuild the cache if the data files
                are modified or if the field options change. Defaults to False.
            train_pca (bool): Whether to compute PCA for each attribute as part of
                dataset metadata compuatation. Defaults to False.
            filter_pred (callable or None): Use only examples for which
                filter_pred(example) is True, or use all examples if None.
                Default is None. This is a keyword-only parameter.

        Returns:
            Tuple[MatchingDataset]: Datasets for (train, validation, and test) splits in
                that order, if provided.
        """

        fields_dict = dict(fields)
        state_args = {'train_pca': train_pca}

        datasets = None
        if cache:
            datafiles = list(f for f in (train, validation, test) if f is not None)
            datafiles = [os.path.expanduser(os.path.join(path, d)) for d in datafiles]
            cachefile = os.path.expanduser(os.path.join(path, cache))
            try:
                cached_data, cache_stale_cause = MatchingDataset.load_cache(
                    fields_dict, datafiles, cachefile, column_naming, state_args)

                if check_cached_data and cache_stale_cause:
                    if not auto_rebuild_cache:
                        raise MatchingDataset.CacheStaleException(cache_stale_cause)
                    else:
                        logger.warning('Rebuilding data cache because: %s', list(cache_stale_cause))

                if not check_cached_data or not cache_stale_cause:
                    datasets = MatchingDataset.restore_data(fields, cached_data)

            except IOError:
                pass

        if not datasets:
            begin = timer()
            dataset_args = {'fields': fields, 'column_naming': column_naming, **kwargs}
            train_data = None if train is None else cls(
                path=os.path.join(path, train), **dataset_args)
            val_data = None if validation is None else cls(
                path=os.path.join(path, validation), **dataset_args)
            test_data = None if test is None else cls(
                path=os.path.join(path, test), **dataset_args)
            datasets = tuple(
                d for d in (train_data, val_data, test_data) if d is not None)

            after_load = timer()
            logger.info('Data load took: {}s'.format(after_load - begin))

            fields_set = set(fields_dict.values())
            for field in fields_set:
                if field is not None and field.use_vocab:
                    field.build_vocab(
                        *datasets, vectors=embeddings, cache=embeddings_cache)
            after_vocab = timer()
            logger.info('Vocab construction time: {}s'.format(after_vocab - after_load))

            if train:
                datasets[0].compute_metadata(train_pca)
            after_metadata = timer()
            logger.info(
                'Metadata computation time: {}s'.format(after_metadata - after_vocab))

            if cache:
                MatchingDataset.save_cache(datasets, fields_dict, datafiles, cachefile,
                                           column_naming, state_args)
                after_cache = timer()
                logger.info('Cache save time: {}s'.format(after_cache - after_vocab))

        if train:
            datasets[0].finalize_metadata()

            # Save additional information to train dataset.
            datasets[0].embeddings = embeddings
            datasets[0].embeddings_cache = embeddings_cache
            datasets[0].train_pca = train_pca

        # Set vocabs.
        for dataset in datasets:
            dataset.vocabs = {
                name: datasets[0].fields[name].vocab
                for name in datasets[0].all_text_fields
            }

        if len(datasets) == 1:
            return datasets[0]
        return tuple(datasets)


def interleave_keys(keys):
    r"""Interleave bits from two sort keys to form a joint sort key.

    Examples that are similar in both of the provided keys will have similar
    values for the key defined by this function. Useful for tasks with two
    text fields like machine translation or natural language inference.
    """

    def interleave(args):
        return ''.join([x for t in zip(*args) for x in t])

    return int(''.join(interleave(format(x, '016b') for x in keys)), base=2)



import pdb

import six

import deepmatcher as dm
import torch
import torch.nn as nn
from torch.autograd import Variable

from . import _utils
from ..batch import AttrTensor


class Pool(dm.modules.Pool, dm.WordAggregator):
    """Pooling based Word Aggregator.

    Takes the same parameters as the :class:`~deepmatcher.modules.Pool` module.
    """
    pass


class AttentionWithRNN(dm.WordAggregator):
    r"""__init__(hidden_size=None, input_dropout=0, rnn='gru', rnn_pool_style='birnn-last', score_dropout=0, input_context_comparison_network='1-layer-highway', value_transform_network=None, transform_dropout=0, input_size=None)

    Attention and RNN based Word Aggregator. This class can be used when the
    aggregation on the primary input also needs the information from the context.
    Specifically, the :class:`~deepmatcher.attr_summarizers.Hybrid` attribute summarizer
    uses this aggregation approach by default. This module takes a primary input sequence
    and a context input sequence and computes a single summary vector for the primary
    input sequence based on the information in the context input. To do so, it does the
    following:

    1. Applies an :ref:`rnn-op` over the context input.
    2. Uses a :ref:`pool-op` operation over the RNN to obtain a single vector summarizing
       the information in the context input.
    3. Based on this context summary vector, uses attention to score the relevance of each
       vector in the primary input sequence.
    4. Performs a weighted average of the vectors in the primary input sequence based on
       the computed scores to obtain a context dependent summary of the primary input
       sequence.

    Args:
        hidden_size (int):
            The default hidden size to use for the RNN,
            `input_context_comparison_network`, and the `value_transform_network`.
        input_dropout (float):
            If non-zero, applies dropout to the input to this module. Dropout probability
            must be between 0 and 1.
        rnn (string or :class:`~deepmatcher.modules.RNN` or callable):
            The RNN used in Step 1 described above. Argument must specify an
            :ref:`rnn-op` operation.
        rnn_pool_style (string):
            The pooling operation used in Step 2 described above. Argument must specify a
            :ref:`pool-op` operation.
        score_dropout (float):
            If non-zero, applies dropout to the attention scores computed in Step 3
            described above. Dropout probability must be between 0 and 1.
        input_context_comparison_network (string or :class:`~deepmatcher.modules.Transform` or callable):
            The neural network that takes each vector in the primary input sequence
            concatenated with the context summary vector to obtain a hidden vector
            representing the primary vector's relevance to the context input.
            Argument must specify a :ref:`transform-op` operation.
        value_transform_network (string or :class:`~deepmatcher.modules.Transform` or callable):
            The neural network to transform the primary input sequence before taking its
            weighted average in Step 4 described above. Argument must be None or specify a
            :ref:`transform-op` operation.
        transform_dropout (float):
            If non-zero, applies dropout to the output of the `value_transform_network`,
            if applicable. Dropout probability must be between 0 and 1.
        input_size (int):
            The number of features in the input to the module. This parameter will be
            automatically specified by :class:`LazyModule`.
    """

    def _init(self,
              hidden_size=None,
              input_dropout=0,
              rnn='gru',
              rnn_pool_style='birnn-last',
              score_dropout=0,
              input_context_comparison_network='1-layer-highway',
              value_transform_network=None,
              transform_dropout=0,
              input_size=None):

        # self.alignment_network = dm.modules._alignment_module(
        #     alignment_network, hidden_size=hidden_size)

        assert rnn is not None
        self.rnn = _utils.get_module(dm.modules.RNN, rnn, hidden_size=hidden_size)
        self.rnn.expect_signature('[AxBxC] -> [AxBx{D}]'.format(D=hidden_size))

        self.rnn_pool = dm.modules.Pool(rnn_pool_style)

        self.input_context_comparison_network = dm.modules._transform_module(
            input_context_comparison_network, hidden_size=hidden_size)
        self.scoring_network = dm.modules._transform_module('1-layer', hidden_size=1)
        self.value_transform_network = dm.modules._transform_module(
            value_transform_network, hidden_size=hidden_size)

        self.input_dropout = nn.Dropout(input_dropout)
        self.transform_dropout = nn.Dropout(transform_dropout)
        self.score_dropout = nn.Dropout(score_dropout)

        self.softmax = nn.Softmax(dim=1)

    def _forward(self, input_with_meta, context_with_meta):
        r"""
        The forward function of the attention-with-RNN netowrk.

        Args:
        input_with_meta ():
            The input sequence with metadata information.
        context_with_meta ():
            The context sequence with metadata information.
        """
        input = self.input_dropout(input_with_meta.data)
        context = self.input_dropout(context_with_meta.data)

        context_rnn_output = self.rnn(
            AttrTensor.from_old_metadata(context, context_with_meta))

        # Dims: batch x 1 x hidden_size
        context_pool_output = self.rnn_pool(context_rnn_output).data.unsqueeze(1)

        # Dims: batch x len1 x hidden_size
        context_pool_repeated = context_pool_output.repeat(1, input.size(1), 1)

        # Dims: batch x len1 x (hidden_size * 2)
        concatenated = torch.cat((input, context_pool_repeated), dim=2)

        # Dims: batch x len1
        raw_scores = self.scoring_network(
            self.input_context_comparison_network(concatenated)).squeeze(2)

        alignment_scores = self.score_dropout(raw_scores)

        if input_with_meta.lengths is not None:
            mask = _utils.sequence_mask(input_with_meta.lengths)
            alignment_scores.data.masked_fill_(~mask, -float('inf'))

        # Make values along dim 2 sum to 1.
        normalized_scores = self.softmax(alignment_scores)

        transformed = input
        if self.value_transform_network is not None:
            transformed = self.transform_dropout(self.value_transform_network(input))

        weighted_sum = torch.bmm(normalized_scores.unsqueeze(1), transformed).squeeze(1)
        return AttrTensor.from_old_metadata(weighted_sum, input_with_meta)



import os
import shutil
import unittest
from collections import Counter
from test import test_dir_path

from nose.tools import *

import torch
from deepmatcher.data.dataset import MatchingDataset
from deepmatcher.data.field import (FastText, FastTextBinary, MatchingField,
                                    MatchingVocab, reset_vector_cache)
from torchtext.vocab import Vectors
from urllib.parse import urljoin
from urllib.request import pathname2url

# import nltk
# nltk.download('perluniprops')
# nltk.download('nonbreaking_prefixes')


class ClassFastTextTestCases(unittest.TestCase):

    def test_init_1(self):
        vectors_cache_dir = '.cache'
        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        pathdir = os.path.abspath(os.path.join(test_dir_path, 'test_datasets'))
        filename = 'fasttext_sample.vec.zip'
        url_base = urljoin('file:', pathname2url(pathdir)) + os.path.sep
        mft = FastText(filename, url_base=url_base, cache=vectors_cache_dir)
        self.assertEqual(mft.dim, 300)
        self.assertEqual(mft.vectors.size(), torch.Size([100, 300]))

        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)


class ClassFastTextBinaryTestCases(unittest.TestCase):

    @raises(RuntimeError)
    def test_init_1(self):
        vectors_cache_dir = '.cache'
        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        pathdir = os.path.abspath(os.path.join(test_dir_path, 'test_datasets'))
        filename = 'fasttext_sample.vec.zip'
        url_base = urljoin('file:', pathname2url(os.path.join(pathdir, filename)))
        mftb = FastTextBinary(filename, url_base=url_base, cache=vectors_cache_dir)

        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

    @raises(OSError)
    def test_init_2(self):
        vectors_cache_dir = '.cache'
        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        pathdir = os.path.abspath(os.path.join(test_dir_path, 'test_datasets'))
        filename = 'fasttext_sample_not_exist.vec.zip'
        url_base = urljoin('file:', pathname2url(os.path.join(pathdir, filename)))
        mftb = FastTextBinary(filename, url_base=url_base, cache=vectors_cache_dir)

        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

    @raises(OSError)
    def test_init_3(self):
        vectors_cache_dir = '.cache'
        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        pathdir = os.path.abspath(os.path.join(test_dir_path, 'test_datasets'))
        filename = 'fasttext_sample_not_exist.gz'
        url_base = urljoin('file:', pathname2url(os.path.join(pathdir, filename)))
        mftb = FastTextBinary(filename, url_base=url_base, cache=vectors_cache_dir)

        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)


class ClassMatchingFieldTestCases(unittest.TestCase):

    def test_init_1(self):
        mf = MatchingField()
        self.assertTrue(mf.sequential)

    def test_init_2(self):
        mf = MatchingField()
        seq = 'Hello, This is a test sequence for tokenizer.'
        tok_seq = [
            'Hello', ',', 'This', 'is', 'a', 'test', 'sequence', 'for', 'tokenizer', '.'
        ]
        self.assertEqual(mf.tokenize(seq), tok_seq)

    @raises(ValueError)
    def test_init_3(self):
        mf = MatchingField(tokenize='random string')

    def test_preprocess_args_1(self):
        mf = MatchingField()
        arg_dict = mf.preprocess_args()
        res_dict = {
            'sequential': True,
            'init_token': None,
            'eos_token': None,
            'init_token': None,
            'lower': False,
            'preprocessing': None,
            'sequential': True,
            'tokenizer_arg': 'nltk',
            'unk_token': '<unk>'
        }
        self.assertEqual(arg_dict, res_dict)

    def test_build_vocab_1(self):
        mf = MatchingField()
        mf.build_vocab()

    @raises(KeyError)
    def test_build_vocab_2(self):
        mf = MatchingField()
        vector_file_name = 'fasttext.wiki_test.vec'
        cache_dir = os.path.join(test_dir_path, 'test_datasets')
        vec_data = mf.build_vocab(vectors=vector_file_name, cache=cache_dir)

    @raises(KeyError)
    def test_build_vocab_3(self):
        mf = MatchingField()
        vector_file_name = 'fasttext.crawl_test.vec'
        cache_dir = os.path.join(test_dir_path, 'test_datasets')
        vec_data = mf.build_vocab(vectors=vector_file_name, cache=cache_dir)
        self.assertIsNone(vec_data)

    def test_get_vector_data(self):
        vectors_cache_dir = '.cache'
        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        pathdir = os.path.abspath(os.path.join(test_dir_path, 'test_datasets'))
        filename = 'fasttext_sample.vec'
        file = os.path.join(pathdir, filename)
        url_base = urljoin('file:', pathname2url(file))
        vecs = Vectors(name=filename, cache=vectors_cache_dir, url=url_base)
        self.assertIsInstance(vecs, Vectors)

        vec_data = MatchingField._get_vector_data(vecs, vectors_cache_dir)
        self.assertEqual(len(vec_data), 1)
        self.assertEqual(vec_data[0].vectors.size(), torch.Size([100, 300]))
        self.assertEqual(vec_data[0].dim, 300)

        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

    def test_numericalize_1(self):
        mf = MatchingField(id=True)
        arr = [[1], [2], [3]]
        mf.numericalize(arr)
        self.assertEqual(arr, [[1], [2], [3]])

    @raises(AttributeError)
    def test_numericalize_2(self):
        mf = MatchingField()
        arr = [['a'], ['b'], ['c']]
        mf.numericalize(arr)

    def test_extend_vocab_1(self):
        vectors_cache_dir = '.cache'
        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        mf = MatchingField()
        lf = MatchingField(id=True, sequential=False)
        fields = [('id', lf), ('left_a', mf), ('right_a', mf), ('label', lf)]
        col_naming = {'id': 'id', 'label': 'label', 'left': 'left_', 'right': 'right_'}

        pathdir = os.path.abspath(os.path.join(test_dir_path, 'test_datasets'))
        filename = 'fasttext_sample.vec'
        file = os.path.join(pathdir, filename)
        url_base = urljoin('file:', pathname2url(file))
        vecs = Vectors(name=filename, cache=vectors_cache_dir, url=url_base)

        data_path = os.path.join(test_dir_path, 'test_datasets', 'sample_table_small.csv')
        md = MatchingDataset(fields, col_naming, path=data_path)

        mf.build_vocab()
        mf.vocab.vectors = torch.Tensor(len(mf.vocab.itos), 300)
        mf.extend_vocab(md, vectors=vecs)
        self.assertEqual(len(mf.vocab.itos), 6)
        self.assertEqual(mf.vocab.vectors.size(), torch.Size([6, 300]))


class TestResetVectorCache(unittest.TestCase):

    def test_reset_vector_cache_1(self):
        mf = MatchingField()
        reset_vector_cache()
        self.assertDictEqual(mf._cached_vec_data, {})


class ClassMatchingVocabTestCases(unittest.TestCase):

    def test_extend_vectors_1(self):
        vectors_cache_dir = '.cache'
        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        pathdir = os.path.abspath(os.path.join(test_dir_path, 'test_datasets'))
        filename = 'fasttext_sample.vec'
        file = os.path.join(pathdir, filename)
        url_base = urljoin('file:', pathname2url(file))
        vecs = Vectors(name=filename, cache=vectors_cache_dir, url=url_base)
        self.assertIsInstance(vecs, Vectors)

        vec_data = MatchingField._get_vector_data(vecs, vectors_cache_dir)
        v = MatchingVocab(Counter())
        v.vectors = torch.Tensor(1, vec_data[0].dim)
        v.unk_init = torch.Tensor.zero_
        tokens = {'hello', 'world'}
        v.extend_vectors(tokens, vec_data)
        self.assertEqual(len(v.itos), 4)
        self.assertEqual(v.vectors.size(), torch.Size([4, 300]))
        self.assertEqual(list(v.vectors[2][0:10]), [0.0] * 10)
        self.assertEqual(list(v.vectors[3][0:10]), [0.0] * 10)

        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)



import io
import os
import shutil
import unittest
from test import test_dir_path

from nose.tools import *

from deepmatcher.data.field import FastText
from deepmatcher.data.process import _check_header, _make_fields, process, process_unlabeled
from torchtext.utils import unicode_csv_reader
from deepmatcher import MatchingModel
from urllib.parse import urljoin
from urllib.request import pathname2url


class CheckHeaderTestCases(unittest.TestCase):

    def test_check_header_1(self):
        path = os.path.join(test_dir_path, 'test_datasets')
        a_dataset = 'sample_table_small.csv'
        with io.open(
                os.path.expanduser(os.path.join(path, a_dataset)), encoding="utf8") as f:
            header = next(unicode_csv_reader(f))
        self.assertEqual(header, ['id', 'left_a', 'right_a', 'label'])
        id_attr = 'id'
        label_attr = 'label'
        left_prefix = 'left'
        right_prefix = 'right'
        _check_header(header, id_attr, left_prefix, right_prefix, label_attr, [])

    @raises(ValueError)
    def test_check_header_2(self):
        path = os.path.join(test_dir_path, 'test_datasets')
        a_dataset = 'sample_table_small.csv'
        with io.open(
                os.path.expanduser(os.path.join(path, a_dataset)), encoding="utf8") as f:
            header = next(unicode_csv_reader(f))
        self.assertEqual(header, ['id', 'left_a', 'right_a', 'label'])
        id_attr = 'id'
        label_attr = 'label'
        left_prefix = 'left'
        right_prefix = 'bb'
        _check_header(header, id_attr, left_prefix, right_prefix, label_attr, [])

    @raises(ValueError)
    def test_check_header_3(self):
        path = os.path.join(test_dir_path, 'test_datasets')
        a_dataset = 'sample_table_small.csv'
        with io.open(
                os.path.expanduser(os.path.join(path, a_dataset)), encoding="utf8") as f:
            header = next(unicode_csv_reader(f))
        self.assertEqual(header, ['id', 'left_a', 'right_a', 'label'])
        id_attr = 'id'
        label_attr = 'label'
        left_prefix = 'aa'
        right_prefix = 'right'
        _check_header(header, id_attr, left_prefix, right_prefix, label_attr, [])

    @raises(ValueError)
    def test_check_header_5(self):
        path = os.path.join(test_dir_path, 'test_datasets')
        a_dataset = 'sample_table_small.csv'
        with io.open(
                os.path.expanduser(os.path.join(path, a_dataset)), encoding="utf8") as f:
            header = next(unicode_csv_reader(f))
        self.assertEqual(header, ['id', 'left_a', 'right_a', 'label'])
        id_attr = 'id'
        label_attr = ''
        left_prefix = 'left'
        right_prefix = 'right'
        _check_header(header, id_attr, left_prefix, right_prefix, label_attr, [])

    @raises(AssertionError)
    def test_check_header_6(self):
        path = os.path.join(test_dir_path, 'test_datasets')
        a_dataset = 'sample_table_small.csv'
        with io.open(
                os.path.expanduser(os.path.join(path, a_dataset)), encoding="utf8") as f:
            header = next(unicode_csv_reader(f))
        self.assertEqual(header, ['id', 'left_a', 'right_a', 'label'])
        header.pop(1)
        id_attr = 'id'
        label_attr = 'label'
        left_prefix = 'left'
        right_prefix = 'right'
        _check_header(header, id_attr, left_prefix, right_prefix, label_attr, [])


class MakeFieldsTestCases(unittest.TestCase):

    def test_make_fields_1(self):
        path = os.path.join(test_dir_path, 'test_datasets')
        a_dataset = 'sample_table_large.csv'
        with io.open(
                os.path.expanduser(os.path.join(path, a_dataset)), encoding="utf8") as f:
            header = next(unicode_csv_reader(f))
        self.assertEqual(header, [
            '_id', 'ltable_id', 'rtable_id', 'label', 'ltable_Song_Name',
            'ltable_Artist_Name', 'ltable_Price', 'ltable_Released', 'rtable_Song_Name',
            'rtable_Artist_Name', 'rtable_Price', 'rtable_Released'
        ])
        id_attr = '_id'
        label_attr = 'label'
        fields = _make_fields(header, id_attr, label_attr, ['ltable_id', 'rtable_id'],
                              True, 'nltk', True)
        self.assertEqual(len(fields), 12)
        counter = {}
        for tup in fields:
            if tup[1] not in counter:
                counter[tup[1]] = 0
            counter[tup[1]] += 1
        self.assertEqual(sorted(list(counter.values())), [1, 1, 2, 8])


class ProcessTestCases(unittest.TestCase):

    def test_process_1(self):
        vectors_cache_dir = '.cache'
        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        data_dir = os.path.join(test_dir_path, 'test_datasets')
        train_path = 'sample_table_large.csv'
        valid_path = 'sample_table_large.csv'
        test_path = 'sample_table_large.csv'
        cache_file = 'cache.pth'
        cache_path = os.path.join(data_dir, cache_file)
        if os.path.exists(cache_path):
            os.remove(cache_path)

        pathdir = os.path.abspath(os.path.join(test_dir_path, 'test_datasets'))
        filename = 'fasttext_sample.vec.zip'
        url_base = urljoin('file:', pathname2url(pathdir)) + os.path.sep
        ft = FastText(filename, url_base=url_base, cache=vectors_cache_dir)

        process(
            data_dir,
            train=train_path,
            validation=valid_path,
            test=test_path,
            id_attr='_id',
            left_prefix='ltable_',
            right_prefix='rtable_',
            cache=cache_file,
            embeddings=ft,
            embeddings_cache_path='')

        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        if os.path.exists(cache_path):
            os.remove(cache_path)

class ProcessUnlabeledTestCases(unittest.TestCase):
    def test_process_unlabeled_1(self):
        vectors_cache_dir = '.cache'
        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)

        data_cache_path = os.path.join(test_dir_path, 'test_datasets',
            'cacheddata.pth')
        if os.path.exists(data_cache_path):
            os.remove(data_cache_path)

        vec_dir = os.path.abspath(os.path.join(test_dir_path, 'test_datasets'))
        filename = 'fasttext_sample.vec.zip'
        url_base = urljoin('file:', pathname2url(vec_dir)) + os.path.sep
        ft = FastText(filename, url_base=url_base, cache=vectors_cache_dir)

        train, valid, test = process(
            path=os.path.join(test_dir_path, 'test_datasets'),
            train='test_train.csv',
            validation='test_valid.csv',
            test='test_test.csv',
            id_attr='id',
            ignore_columns=('left_id', 'right_id'),
            embeddings=ft,
            embeddings_cache_path='',
            pca=True)

        model_save_path = 'sif_model.pth'
        model = MatchingModel(attr_summarizer='sif')
        model.run_train(
            train,
            valid,
            epochs=1,
            batch_size=8,
            best_save_path= model_save_path,
            pos_neg_ratio=3)

        test_unlabeled = process_unlabeled(
            path=os.path.join(test_dir_path, 'test_datasets', 'test_test.csv'),
            trained_model=model,
            ignore_columns=('left_id', 'right_id'))

        self.assertEqual(test_unlabeled.all_text_fields, test.all_text_fields)

        if os.path.exists(model_save_path):
            os.remove(model_save_path)

        if os.path.exists(data_cache_path):
            os.remove(data_cache_path)

        if os.path.exists(vectors_cache_dir):
            shutil.rmtree(vectors_cache_dir)



import pdb

import six

import deepmatcher as dm
import torch
import torch.nn as nn
from torch.autograd import Variable

from . import _utils
from ..batch import AttrTensor


class Attention(dm.WordComparator):
    r"""__init__(heads=1, hidden_size=None, raw_alignment=False, input_dropout=0, alignment_network='decomposable', scale=False, score_dropout=0, value_transform_network=None, input_transform_network=None, value_merge='concat', transform_dropout=0, comparison_merge='concat', comparison_network='2-layer-highway', input_size=None)

    Attention based Word Comparator with `multi-head <https://arxiv.org/abs/1706.03762>`__
    support. This module does the following:

    1. Computes an alignment matrix between the primary input sequence and the context
       input sequence.
    2. For each vector in the primary input sequence, takes a weighted average over all
       vectors in the context input sequence, where weights are given by the alignment
       matrix. Intuitively, for each word / phrase vector in the primary input sequence,
       this represents the aligning word / phrase vector in the context input sequence.
    3. Compares the vectors in the primary input sequence with its aligning vectors.

    Args:
        heads (int):
            Number of attention heads to use. Defaults to 1.
        hidden_size (int):
            The default hidden size of the `alignment_network`, transform networks (if
            applicable), and comparison network.
        raw_alignment (bool):
            If True, uses the contextualized version (transformed by the Word
            Contextualizer module) of the input and context sequences for computing
            alignment in Step 1 described above. If False, uses the raw
            (non-contextualized) word embedding sequences for computing alignment in Step
            1 described above. For step 2, the Word Contextualizer's version of the
            context sequence is used for computing the weighted averages in both cases.
            Raw alignment has been shown to `perform better and speed up convergence
            <http://pages.cs.wisc.edu/~anhai/papers1/deepmatcher-sigmod18.pdf>`__,
            especially in cases of limited training data.
        input_dropout (float):
            If non-zero, applies dropout to the input to this module. Dropout probability
            must be between 0 and 1.
        alignment_network (string or :class:`deepmatcher.modules.AlignmentNetwork` or callable):
            The neural network that takes the primary input sequence, aligns the word /
            phrase vectors in this sequence with word / phrase vector in the context
            sequence, and returns the corresponding alignment score matrix. Argument must
            specify a :ref:`align-op` operation.
        scale (bool):
            Whether to scale the alignment scores by the square root of the
            `hidden_size` parameter. Based on `scaled dot-product attention
            <https://arxiv.org/abs/1706.03762>`__
        score_dropout (float):
            If non-zero, applies dropout to the alignment score matrix. Dropout
            probability must be between 0 and 1.
        value_transform_network (string or :class:`~deepmatcher.modules.Transform` or callable):
            The neural network to transform the context input sequence before taking the
            weighted averages in Step 2 described above. Argument must be None or specify
            a :ref:`transform-op` operation. If the argument is a string, the hidden size
            of the transform operation is computed as :code:`hidden_size // heads`. If
            argument is None, and `heads` is 1, then the values are not transformed. If
            argument is None and `heads` is > 1, then a 1 layer highway network without
            any non-linearity is used. The hidden size for this is computed as mentioned
            above.
        input_transform_network (string or :class:`~deepmatcher.modules.Transform` or callable):
            The neural network to transform the primary input sequence before it is
            compared with the aligning vectors in the context sequence in Step 3 described
            above. Argument must be None or specify a :ref:`transform-op` operation. If
            None, uses the same neural network as the value transform network (sharing not
            just the structure but also weight parameters).
        value_merge (string or :class:`~deepmatcher.modules.Merge` or callable):
            Specifies how to merge the outputs of all attention heads for each vector
            in the primary input sequence. Concatenates the outputs of all heads by
            default. Argument must specify a :ref:`merge-op` operation.
        transform_dropout (float):
            If non-zero, applies dropout to the outputs of the `value_transform_network`
            and `input_transform_network`, if applicable. Dropout probability must be
            between 0 and 1.
        comparison_merge (string or :class:`~deepmatcher.modules.Merge` or callable):
            For each vector in the primary input sequence, specifies how to merge it with
            its aligning vector in the context input sequence, to obtain a single vector.
            The resulting sequence vectors forms the input to the `comparison_network`.
            Concatenates each primary input vector with its aligning vector by default.
            Argument must specify a :ref:`merge-op` operation.
        comparison_network (string or :class:`~deepmatcher.modules.Transform` or callable):
            The neural network to compare the vectors in the primary input sequence and
            their aligning vectors in the context input. Input to this module is produced
            by the `comparison_merge` operation. Argument must specify a
            :ref:`transform-op` operation.
        input_size (int):
            The number of features in the input to the module. This parameter will be
            automatically specified by :class:`LazyModule`.
    """

    def _init(self,
              heads=1,
              hidden_size=None,
              raw_alignment=False,
              input_dropout=0,
              alignment_network='decomposable',
              scale=False,
              score_dropout=0,
              value_transform_network=None,
              input_transform_network=None,
              value_merge='concat',
              transform_dropout=0,
              comparison_merge='concat',
              comparison_network='2-layer-highway',
              input_size=None):
        hidden_size = hidden_size if hidden_size is not None else input_size[0]

        self.alignment_networks = nn.ModuleList()
        for head in range(heads):
            self.alignment_networks.append(
                dm.modules._alignment_module(alignment_network, hidden_size))

        if value_transform_network is None and heads > 1:
            value_transform_network = dm.modules.Transform(
                '1-layer-highway', non_linearity=None, hidden_size=hidden_size // heads)
        self.value_transform_network = dm.modules._transform_module(
            value_transform_network, hidden_size // heads)

        if input_transform_network is None:
            self.input_transform_network = self.value_transform_network
        else:
            self.input_transform_network = dm.modules._transform_module(
                input_transform_network, hidden_size // heads)

        self.value_merge = dm.modules._merge_module(value_merge)
        self.comparison_merge = dm.modules._merge_module(comparison_merge)
        self.comparison_network = dm.modules._transform_module(comparison_network,
                                                               hidden_size)

        self.input_dropout = nn.Dropout(input_dropout)
        self.transform_dropout = nn.Dropout(transform_dropout)
        self.score_dropout = nn.Dropout(score_dropout)
        self.softmax = nn.Softmax(dim=2)

        self.raw_alignment = raw_alignment
        self.heads = heads
        self.scale = scale

    def _forward(self,
                 input_with_meta,
                 context_with_meta,
                 raw_input_with_meta=None,
                 raw_context_with_meta=None):
        input = self.input_dropout(input_with_meta.data)
        context = self.input_dropout(context_with_meta.data)
        raw_input = self.input_dropout(raw_input_with_meta.data)
        raw_context = self.input_dropout(raw_context_with_meta.data)

        queries = input
        keys = context
        values = context
        if self.raw_alignment:
            queries = raw_input
            keys = raw_context

        inputs_transformed = []
        values_aligned = []
        for head in range(self.heads):
            # Dims: batch x len1 x len2
            alignment_scores = self.score_dropout(self.alignment_networks[head](queries,
                                                                                keys))
            if self.scale:
                alignment_scores = alignment_scores / torch.sqrt(hidden_size)

            if context_with_meta.lengths is not None:
                mask = _utils.sequence_mask(context_with_meta.lengths)
                mask = mask.unsqueeze(1)  # Make it broadcastable.
                alignment_scores.data.masked_fill_(~mask, -float('inf'))

            # Make values along dim 2 sum to 1.
            normalized_scores = self.softmax(alignment_scores)

            if self.input_transform_network is not None:
                inputs_transformed.append(
                    self.transform_dropout(self.input_transform_network(input)))
            if self.value_transform_network is not None:
                values_transformed = self.transform_dropout(
                    self.value_transform_network(values))
            else:
                values_transformed = values

            # Dims: batch x len1 x channels
            values_aligned.append(torch.bmm(normalized_scores, values_transformed))

        inputs_merged = input
        if inputs_transformed:
            inputs_merged = self.value_merge(*inputs_transformed)
        values_merged = self.value_merge(*values_aligned)

        comparison_merged = self.comparison_merge(inputs_merged, values_merged)
        comparison = self.comparison_network(comparison_merged)

        return AttrTensor.from_old_metadata(comparison, input_with_meta)



import six

import deepmatcher as dm
import torch
import torch.nn as nn
from torch.autograd import Variable

from . import _utils
from ..batch import AttrTensor


class RNN(dm.modules.RNN, dm.WordContextualizer):
    """Multi layered RNN based Word Contextualizer.

    Supports dropout and residual / highway connections. Takes the same parameters as the
    :class:`~deepmatcher.modules.RNN` module.
    """
    pass


# class CNN(dm.WordContextualizer):
#     pass


class SelfAttention(dm.WordContextualizer):
    """__init__(heads=1, hidden_size=None, input_dropout=0, alignment_network='decomposable', scale=False, score_dropout=0, value_transform_network=None, value_merge='concat', transform_dropout=0, output_transform_network=None, output_dropout=0, bypass_network='highway', input_size=None)

    Self Attention based Word Contextualizer.

    Supports `vanilla self attention <https://arxiv.org/abs/1606.01933>`__ and `multi-head
    self attention <https://arxiv.org/abs/1706.03762>`__.

    Args:
        heads (int):
            Number of attention heads to use. Defaults to 1.
        hidden_size (int):
            The default hidden size of the `alignment_network` and transform networks, if
            they are not disabled.
        input_dropout (float):
            If non-zero, applies dropout to the input to this module. Dropout probability
            must be between 0 and 1.
        alignment_network (string or :class:`deepmatcher.modules.AlignmentNetwork` or callable):
            The neural network takes the input sequence, aligns the words in the sequence
            with other words in the sequence, and returns the corresponding alignment
            score matrix. Argument must specify a :ref:`align-op` operation.
        scale (bool):
            Whether to scale the alignment scores by the square root of the
            `hidden_size` parameter. Based on `scaled dot-product attention
            <https://arxiv.org/abs/1706.03762>`__
        score_dropout (float):
            If non-zero, applies dropout to the alignment score matrix. Dropout
            probability must be between 0 and 1.
        value_transform_network (string or :class:`~deepmatcher.modules.Transform` or callable):
            For each word embedding in the input sequence, SelfAttention takes a weighted
            average of the aligning values, i.e., the aligning word embeddings based on
            the alignment scores. This parameter specifies the neural network to transform
            the values (word embeddings) before taking the weighted average. Argument must
            be None or specify a :ref:`transform-op` operation. If the argument is a
            string, the hidden size of the transform operation is computed as
            :code:`hidden_size // heads`. If argument is None, and `heads` is 1, then the
            values are not transformed. If argument is None and `heads` is > 1, then a 1
            layer highway network without any non-linearity is used. The hidden size for
            this is computed as mentioned above.
        value_merge (string or :class:`~deepmatcher.modules.Merge` or callable):
            For each word embedding in the input sequence, each SelfAttention head
            produces one corresponding vector as output. This parameter specifies how
            to merge the outputs of all attention heads for each word embedding.
            Concatenates the outputs of all heads by default. Argument must specify a
            :ref:`merge-op` operation.
        transform_dropout (float):
            If non-zero, applies dropout to the output of the `value_transform_network`,
            if applicable. Dropout probability must be between 0 and 1.
        output_transform_network (string or :class:`~deepmatcher.modules.Transform` or callable):
            For each word embedding in the input sequence, SelfAttention produces one
            corresponding vector as output. This neural network specifies how to transform
            each of these output vectors to obtain a hidden representation of size
            `hidden_size`. Argument must be None or specify a :ref:`transform-op`
            operation. If argument is None, and `heads` is 1, then the output vectors are
            not transformed. If argument is None and `heads` is > 1, then a 1 layer
            highway network without any non-linearity is used.
        output_dropout (float):
            If non-zero, applies dropout to the output of the `output_transform_network`,
            if applicable. Dropout probability must be between 0 and 1.
        bypass_network (string or :class:`Bypass` or callable):
            The bypass network (e.g. residual or highway network) to use. The input word
            embedding sequence to this module is considered as the raw input to the bypass
            network and the final output vector sequence (output of `value_merge` or
            `output_transform_network` if applicable) is considered as the transformed
            input. Argument must specify a :ref:`bypass-op` operation. If None, does not
            use a bypass network.
        input_size (int):
            The number of features in the input to the module. This parameter will be
            automatically specified by :class:`LazyModule`.
    """

    def _init(self,
              heads=1,
              hidden_size=None,
              input_dropout=0,
              alignment_network='decomposable',
              scale=False,
              score_dropout=0,
              value_transform_network=None,
              value_merge='concat',
              transform_dropout=0,
              output_transform_network=None,
              output_dropout=0,
              bypass_network='highway',
              input_size=None):
        hidden_size = hidden_size if hidden_size is not None else input_size

        self.alignment_networks = nn.ModuleList()
        for head in range(heads):
            self.alignment_networks.append(
                dm.modules._alignment_module(alignment_network, hidden_size))

        if value_transform_network is None and heads > 1:
            value_transform_network = dm.modules.Transform(
                '1-layer-highway', non_linearity=None, hidden_size=hidden_size // heads)
        self.value_transform_network = dm.modules._transform_module(
            value_transform_network, hidden_size // heads)

        self.value_merge = dm.modules._merge_module(value_merge)

        self.softmax = nn.Softmax(dim=2)

        if output_transform_network is None and heads > 1:
            output_transform_network = dm.modules.Transform(
                '1-layer-highway', non_linearity=None, hidden_size=hidden_size)
        self.output_transform_network = dm.modules._transform_module(
            output_transform_network, hidden_size)

        self.input_dropout = nn.Dropout(input_dropout)
        self.transform_dropout = nn.Dropout(transform_dropout)
        self.score_dropout = nn.Dropout(output_dropout)
        self.output_dropout = nn.Dropout(output_dropout)

        self.bypass_network = dm.modules._bypass_module(bypass_network)

        self.heads = heads
        self.scale = scale
        self.hidden_size = hidden_size

    def _forward(self, input_with_meta):
        input = self.input_dropout(input_with_meta.data)

        values_aligned = []
        for head in range(self.heads):
            # Dims: batch x len1 x len2
            alignment_scores = self.score_dropout(self.alignment_networks[head](input,
                                                                                input))

            if self.scale:
                alignment_scores = alignment_scores / torch.sqrt(self.hidden_size)

            if input_with_meta.lengths is not None:
                mask = _utils.sequence_mask(input_with_meta.lengths)
                mask = mask.unsqueeze(1)  # Make it broadcastable.
                alignment_scores.data.masked_fill_(~mask, -float('inf'))

            normalized_scores = self.softmax(alignment_scores)

            if self.value_transform_network is not None:
                values_transformed = self.transform_dropout(
                    self.value_transform_network(input))
            else:
                values_transformed = input

            # Dims: batch x len1 x channels
            values_aligned.append(torch.bmm(normalized_scores, values_transformed))

        values_merged = self.value_merge(*values_aligned)

        output = values_merged
        if self.output_transform_network:
            output = self.output_transform_network(output)
        output = self.output_dropout(output)

        final_output = self.bypass_network(output, input)

        return AttrTensor.from_old_metadata(final_output, input_with_meta)



from __future__ import division

import logging

from torchtext.legacy import data

from ..batch import MatchingBatch

logger = logging.getLogger(__name__)


class MatchingIterator(data.BucketIterator):

    def __init__(self,
                 dataset,
                 train_info,
                 train,
                 batch_size,
                 sort_in_buckets=None,
                 device=None,
                 **kwargs):
        if sort_in_buckets is None:
            sort_in_buckets = train
        self.sort_in_buckets = sort_in_buckets
        self.train_info = train_info
        super(MatchingIterator, self).__init__(
            dataset, batch_size, train=train, repeat=False, sort=False, device=device, **kwargs)

    @classmethod
    def splits(cls, datasets, batch_sizes=None, **kwargs):
        """Create Iterator objects for multiple splits of a dataset.

        Args:
            datasets: Tuple of Dataset objects corresponding to the splits. The
                first such object should be the train set.
            batch_sizes: Tuple of batch sizes to use for the different splits,
                or None to use the same batch_size for all splits.
            Remaining keyword arguments: Passed to the constructor of the
                iterator class being used.
        """
        if batch_sizes is None:
            batch_sizes = [kwargs.pop('batch_size')] * len(datasets)
        ret = []
        for i in range(len(datasets)):
            ret.append(
                cls(datasets[i],
                    train_info=datasets[0],
                    train=i==0,
                    batch_size=batch_sizes[i],
                    **kwargs))
        return tuple(ret)

    def __iter__(self):
        for batch in super(MatchingIterator, self).__iter__():
            yield MatchingBatch(batch, self.train_info)

    def create_batches(self):
        if self.sort_in_buckets:
            return data.BucketIterator.create_batches(self)
        else:
            return data.Iterator.create_batches(self)



"""Defines built-in attribute summarizers."""

import deepmatcher as dm


class SIF(dm.AttrSummarizer):
    """__init__(word_contextualizer=None, word_comparator=None, word_aggregator=None, \
        hidden_size=None)

    The attribute summarizer for the SIF (Smooth Inverse Frequency) model.

    Args:
        word_contextualizer (string or :class:`~deepmatcher.WordContextualizer` or callable): The
            word contextualizer module (refer to :class:`~deepmatcher.WordContextualizer` for
            details) to use for attribute summarization. The SIF model does not take word
            context information into account, hence this defaults to None.
        word_comparator (string or :class:`~deepmatcher.WordComparator` or callable): The word
            comparator module (refer to :class:`~deepmatcher.WordComparator` for details) to use
            for attribute summarization. The SIF model does not perform word by word
            comparisons, hence this defaults to None.
        word_aggregator (string or :class:`~deepmatcher.WordAggregator` or callable): The word
            aggregator module (refer to :class:`~deepmatcher.WordAggregator` for details) to use
            for attribute summarization. This model uses SIF-based weighted average
            aggregation over the  word embeddings of an input sequence, hence this
            defaults to 'sif-pool'.
        hidden_size (int): The hidden size to use for all 3 attribute summarization
            sub-modules (i.e., word contextualizer, word comparator, and word aggregator),
            if they are customized. By default, the SIF model does not use this parameter.
    """

    def _init(self,
              word_contextualizer=None,
              word_comparator=None,
              word_aggregator=None,
              hidden_size=None):
        word_aggregator = word_aggregator or 'sif-pool'
        super(SIF, self)._init(
            word_contextualizer=word_contextualizer,
            word_comparator=word_comparator,
            word_aggregator=word_aggregator,
            hidden_size=hidden_size)


class RNN(dm.AttrSummarizer):
    r"""The attribute summarizer for the RNN model.

    Args:
        word_contextualizer (string or :class:`~deepmatcher.WordContextualizer` or callable): The
            word contextualizer module (refer to :class:`~deepmatcher.WordContextualizer` for
            details) to use for attribute summarization. This model uses RNN to take into
            account the context information, and the default value is 'gru' (i.e., uses
            the bidirectional GRU model as the specific RNN instantiation.) Other options
            are 'rnn' (the vanilla bi-RNN) and 'lstm' (the bi-LSTM model).
        word_comparator (string or :class:`~deepmatcher.WordComparator` or callable): The word
            comparator module (refer to :class:`~deepmatcher.WordComparator` for details) to use
            for attribute summarization. The RNN model does not perform word by word
            comparisons, hence this defaults to None.
        word_aggregator (string or :class:`~deepmatcher.WordAggregator` or callable): The word
            aggregator module (refer to :class:`~deepmatcher.WordAggregator` for details) to use
            for attribute summarization. The RNN model uses bi-directional RNN and
            concatenates the last ouputs of the forward and backward RNNs, hence the
            default value is 'birnn-last-pool'.
        hidden_size (int): The hidden size to use for the word contextualizer. This value
            will also be used as the hidden size for the other 2 attribute summarization
            sub-modules (i.e., word comparator, and word aggregator), if they are
            customized. If not specified, the hidden size for each component will be set
            to be the same as its input size. E.g. if the word embedding dimension is 300
            and hidden_size is None, the word contextualizer's hidden size will be 300.
    """

    def _init(self,
              word_contextualizer=None,
              word_comparator=None,
              word_aggregator=None,
              hidden_size=None):
        word_contextualizer = word_contextualizer or 'gru'
        word_aggregator = word_aggregator or 'birnn-last-pool'
        super(RNN, self)._init(
            word_contextualizer=word_contextualizer,
            word_comparator=word_comparator,
            word_aggregator=word_aggregator,
            hidden_size=hidden_size)


class Attention(dm.AttrSummarizer):
    r"""The attribute summarizer for the attention-based model.

    Args:
        word_contextualizer (string or :class:`~deepmatcher.WordContextualizer` or callable): The
            word contextualizer module (refer to :class:`~deepmatcher.WordContextualizer` for
            details) to use for attribute summarization. The attention model does not take
            word context information into account, hence this defaults to None.
        word_comparator (string or :class:`~deepmatcher.WordComparator` or callable): The word
            comparator module (refer to :class:`~deepmatcher.WordComparator` for details) to use
            for attribute summarization. The attention model performs word by word
            comparison with the decomposable attention mechanism, hence this defaults to
            'decomposable-attention'.
        word_aggregator (string or :class:`~deepmatcher.WordAggregator` or callable): The word
            aggregator module (refer to :class:`~deepmatcher.WordAggregator` for details) to use
            for attribute summarization. The Attention model performs the aggregation by
            summing over the comparison results from the word comparator, divided by the
            length of the input sequence (to get constant variance through the network
            flow). Hence this defaults to 'divsqrt-pool'.
        hidden_size (int): The hidden size to use for the word comparator. This value
            will also be used as the hidden size for the other 2 attribute summarization
            sub-modules (i.e., word contextualizer, and word aggregator), if they are
            customized. If not specified, the hidden size for each component will be set
            to be the same as its input size. E.g. if the word embedding dimension is 300
            and hidden_size is None, the word contextualizer's hidden size will be 300.
    """

    def _init(self,
              word_contextualizer=None,
              word_comparator=None,
              word_aggregator=None,
              hidden_size=None):
        word_comparator = word_comparator or 'decomposable-attention'
        word_aggregator = word_aggregator or 'divsqrt-pool'
        super(Attention, self)._init(
            word_contextualizer=word_contextualizer,
            word_comparator=word_comparator,
            word_aggregator=word_aggregator,
            hidden_size=hidden_size)


class Hybrid(dm.AttrSummarizer):
    r"""The attribute summarizer for the hybrid model.

    Args:
        word_contextualizer (string or :class:`~deepmatcher.WordContextualizer` or callable): The
            word contextualizer module (refer to :class:`~deepmatcher.WordContextualizer` for
            details) to use for attribute summarization. The hybrid model uses
            bidirectional GRU(a specific type of RNN) to take into account the context
            information. The default value is 'gru'.
        word_comparator (string or :class:`~deepmatcher.WordComparator` or callable): The word
            comparator module (refer to :class:`~deepmatcher.WordComparator` for details) to use
            for attribute summarization. The hybrid model performs word by word comparison
            over the raw input word embeddings (rather than the RNN hiddens states), hence
            this defaults to an Attention object with 'decomposable' as the attention
            mechanism on the raw input embeddings.
        word_aggregator (string or :class:`~deepmatcher.WordAggregator` or callable): The word
            aggregator module (refer to :class:`~deepmatcher.WordAggregator` for details) to use
            for attribute summarization. A second layer of attention has been used for the
            aggregation. Please consult the paper for more information. The default value
            is 'concat-attention-with-rnn'.
        hidden_size (int): The hidden size to use for all 3 attribute summarization
            sub-modules (i.e., word contextualizer, word comparator, and word aggregator),
            if they are customized.
    """

    def _init(self,
              word_contextualizer=None,
              word_comparator=None,
              word_aggregator=None,
              hidden_size=None):
        word_contextualizer = word_contextualizer or 'gru'
        word_comparator = word_comparator or dm.word_comparators.Attention(
            alignment_network='decomposable', raw_alignment=True)
        word_aggregator = word_aggregator or 'attention-with-rnn'
        super(Hybrid, self)._init(
            word_contextualizer=word_contextualizer,
            word_comparator=word_comparator,
            word_aggregator=word_aggregator,
            hidden_size=hidden_size)



import copy
import io
import logging
import os
from timeit import default_timer as timer

import six

from torchtext.utils import unicode_csv_reader

from .dataset import MatchingDataset
from .field import MatchingField

logger = logging.getLogger(__name__)


def _check_header(header, id_attr, left_prefix, right_prefix, label_attr, ignore_columns):
    r"""Verify CSV file header.

    Checks that:
    * There is a label column
    * There is an ID column
    * All columns except the label and ID columns, and ignored columns start with either
        the left table attribute prefix or the right table attribute prefix.
    * The number of left and right table attributes are the same.
    """
    # assert id_attr in header
    if label_attr:
        assert label_attr in header

    for attr in header:
        if attr not in (id_attr, label_attr) and attr not in ignore_columns:
            if not attr.startswith(left_prefix) and not attr.startswith(right_prefix):
                raise ValueError('Attribute ' + attr + ' is not a left or a right table '
                                 'column, not a label or id and is not ignored. Not sure '
                                 'what it is...')

    num_left = sum(attr.startswith(left_prefix) for attr in header)
    num_right = sum(attr.startswith(right_prefix) for attr in header)
    assert num_left == num_right


def _make_fields(header, id_attr, label_attr, ignore_columns, lower, tokenize,
                 include_lengths):
    r"""Create field metadata, i.e., attribute processing specification for each
    attribute.

    This includes fields for label and ID columns.

    Returns:
        list(tuple(str, MatchingField)): A list of tuples containing column name
            (e.g. "left_address") and corresponding :class:`~data.MatchingField` pairs,
            in the same order that the columns occur in the CSV file.
    """

    text_field = MatchingField(
        lower=lower,
        tokenize=tokenize,
        init_token='<<<',
        eos_token='>>>',
        batch_first=True,
        include_lengths=include_lengths)
    numeric_field = MatchingField(
        sequential=False, preprocessing=lambda x: int(x), use_vocab=False)
    id_field = MatchingField(sequential=False, use_vocab=False, id=True)

    fields = []
    for attr in header:
        if attr == id_attr:
            fields.append((attr, id_field))
        elif attr == label_attr:
            fields.append((attr, numeric_field))
        elif attr in ignore_columns:
            fields.append((attr, None))
        else:
            fields.append((attr, text_field))
    return fields


def _maybe_download_nltk_data():
    import nltk
    nltk.download('perluniprops', quiet=True)
    nltk.download('nonbreaking_prefixes', quiet=True)
    nltk.download('punkt', quiet=True)


def process(path,
            train=None,
            validation=None,
            test=None,
            unlabeled=None,
            cache='cacheddata.pth',
            check_cached_data=True,
            auto_rebuild_cache=True,
            tokenize='nltk',
            lowercase=True,
            embeddings='fasttext.en.bin',
            embeddings_cache_path='~/.vector_cache',
            ignore_columns=(),
            include_lengths=True,
            id_attr='id',
            label_attr='label',
            left_prefix='left_',
            right_prefix='right_',
            use_magellan_convention=False,
            pca=True):
    """Creates dataset objects for multiple splits of a dataset.

    This involves the following steps (if data cannot be retrieved from the cache):
    #. Read CSV header of a data file and verify header is sane.
    #. Create fields, i.e., column processing specifications (e.g. tokenization, label
        conversion to integers etc.)
    #. Load each data file:
        #. Read each example (tuple pair) in specified CSV file.
        #. Preprocess example. Involves lowercasing and tokenization (unless disabled).
        #. Compute metadata if training data file. \
            See :meth:`MatchingDataset.compute_metadata` for details.
    #. Create vocabulary consisting of all tokens in all attributes in all datasets.
    #. Download word embedding data if necessary.
    #. Create mapping from each word in vocabulary to its word embedding.
    #. Compute metadata
    #. Write to cache

    Arguments:
        path (str): Common prefix of the splits' file paths.
        train (str): Suffix to add to path for the train set.
        validation (str): Suffix to add to path for the validation set, or None
            for no validation set. Default is None.
        test (str): Suffix to add to path for the test set, or None for no test
            set. Default is None.
        cache (str): Suffix to add to path for cache file. If `None` disables caching.
        check_cached_data (bool): Verify that data files haven't changes since the
            cache was constructed and that relevant field options haven't changed.
        auto_rebuild_cache (bool): Automatically rebuild the cache if the data files
            are modified or if the field options change. Defaults to False.
        lowercase (bool): Whether to lowercase all words in all attributes.
        embeddings (str or list): One or more of the following strings:

            * `fasttext.{lang}.bin`:
                This uses sub-word level word embeddings based on binary models from "wiki
                word vectors" released by FastText. {lang} is 'en' or any other 2 letter
                ISO 639-1 Language Code, or 3 letter ISO 639-2 Code, if the language does
                not have a 2 letter code. 300d vectors.
                ``fasttext.en.bin`` is the default.
            * `fasttext.wiki.vec`:
                Uses wiki news word vectors released as part of "Advances in Pre-Training
                Distributed Word Representations" by Mikolov et al. (2018). 300d vectors.
            * `fasttext.crawl.vec`:
                Uses Common Crawl word vectors released as part of "Advances in
                Pre-Training Distributed Word Representations" by Mikolov et al. (2018).
                300d vectors.
            * `glove.6B.{dims}`:
                Uses uncased Glove trained on Wiki + Gigaword. {dims} is one of (50d,
                100d, 200d, or 300d).
            * `glove.42B.300d`:
                Uses uncased Glove trained on Common Crawl. 300d vectors.
            * `glove.840B.300d`:
                Uses cased Glove trained on Common Crawl. 300d vectors.
            * `glove.twitter.27B.{dims}`:
                Uses cased Glove trained on Twitter. {dims} is one of (25d, 50d, 100d, or
                200d).
        embeddings_cache_path (str): Directory to store dowloaded word vector data.
        ignore_columns (list): A list of columns to ignore in the CSV files.
        include_lengths (bool): Whether to provide the model with the lengths of
            each attribute sequence in each batch. If True, length information can be
            used by the neural network, e.g. when picking the last RNN output of each
            attribute sequence.
        id_attr (str): The name of the tuple pair ID column in the CSV file.
        label_attr (str): The name of the tuple pair match label column in the CSV file.
        left_prefix (str): The prefix for attribute names belonging to the left table.
        right_prefix (str): The prefix for attribute names belonging to the right table.
        use_magellan_convention (bool): Set `id_attr`, `left_prefix`, and `right_prefix`
            according to Magellan (py_entitymatching Python package) naming conventions.
            Specifically, set them to be '_id', 'ltable_', and 'rtable_' respectively.
        pca (bool): Whether to compute PCA for each attribute (needed for SIF model).
            Defaults to False.

    Returns:
        Tuple[MatchingDataset]: Datasets for (train, validation, and test) splits in that
            order, if provided, or dataset for unlabeled, if provided.
    """
    if unlabeled is not None:
        raise ValueError('Parameter "unlabeled" has been deprecated, use '
                         '"deepmatcher.data.process_unlabeled" instead.')

    if use_magellan_convention:
        id_attr = '_id'
        left_prefix = 'ltable_'
        right_prefix = 'rtable_'

    # TODO(Sid): check for all datasets to make sure the files exist and have the same schema
    a_dataset = train or validation or test
    with io.open(os.path.expanduser(os.path.join(path, a_dataset)), encoding="utf8") as f:
        header = next(unicode_csv_reader(f))

    _maybe_download_nltk_data()
    _check_header(header, id_attr, left_prefix, right_prefix, label_attr, ignore_columns)
    fields = _make_fields(header, id_attr, label_attr, ignore_columns, lowercase,
                          tokenize, include_lengths)

    column_naming = {
        'id': id_attr,
        'left': left_prefix,
        'right': right_prefix,
        'label': label_attr
    }

    datasets = MatchingDataset.splits(
        path,
        train,
        validation,
        test,
        fields,
        embeddings,
        embeddings_cache_path,
        column_naming,
        cache,
        check_cached_data,
        auto_rebuild_cache,
        train_pca=pca)

    # Save additional information to train dataset.
    datasets[0].ignore_columns = ignore_columns
    datasets[0].tokenize = tokenize
    datasets[0].lowercase = lowercase
    datasets[0].include_lengths = include_lengths

    return datasets


def process_unlabeled(path, trained_model, ignore_columns=None):
    """Creates a dataset object for an unlabeled dataset.

    Args:
        path (string):
            The full path to the unlabeled data file (not just the directory).
        trained_model (:class:`~deepmatcher.MatchingModel`):
            The trained model. The model is aware of the configuration of the training
            data on which it was trained, and so this method reuses the same
            configuration for the unlabeled data.
        ignore_columns (list):
            A list of columns to ignore in the unlabeled CSV file.
    """
    with io.open(path, encoding="utf8") as f:
        header = next(unicode_csv_reader(f))

    train_info = trained_model.meta
    if ignore_columns is None:
        ignore_columns = train_info.ignore_columns
    column_naming = dict(train_info.column_naming)
    column_naming['label'] = None

    fields = _make_fields(header, column_naming['id'], column_naming['label'],
                          ignore_columns, train_info.lowercase, train_info.tokenize,
                          train_info.include_lengths)

    begin = timer()
    dataset_args = {'fields': fields, 'column_naming': column_naming}
    dataset = MatchingDataset(path=path, **dataset_args)

    # Make sure we have the same attributes.
    assert set(dataset.all_text_fields) == set(train_info.all_text_fields)

    after_load = timer()
    logger.info('Data load time: {}s'.format(after_load - begin))

    reverse_fields_dict = dict((pair[1], pair[0]) for pair in fields)
    for field, name in reverse_fields_dict.items():
        if field is not None and field.use_vocab:
            # Copy over vocab from original train data.
            field.vocab = copy.deepcopy(train_info.vocabs[name])
            # Then extend the vocab.
            field.extend_vocab(
                dataset, vectors=train_info.embeddings, cache=train_info.embeddings_cache)

    dataset.vocabs = {
        name: dataset.fields[name].vocab
        for name in train_info.all_text_fields
    }

    after_vocab = timer()
    logger.info('Vocab update time: {}s'.format(after_vocab - after_load))

    return dataset



import logging

import torch
import torch.nn as nn
import torch.optim as optim
from torch.autograd import Variable
from torch.nn.utils import clip_grad_norm_

logger = logging.getLogger('deepmatcher.optim')


class SoftNLLLoss(nn.NLLLoss):
    """A soft version of negative log likelihood loss with support for label smoothing.

    Effectively equivalent to PyTorch's :class:`torch.nn.NLLLoss`, if `label_smoothing`
    set to zero. While the numerical loss values will be different compared to
    :class:`torch.nn.NLLLoss`, this loss results in the same gradients. This is because
    the implementation uses :class:`torch.nn.KLDivLoss` to support multi-class label
    smoothing.

    Args:
        label_smoothing (float):
            The smoothing parameter :math:`epsilon` for label smoothing. For details on
            label smoothing refer `this paper <https://arxiv.org/abs/1512.00567v1>`__.
        weight (:class:`torch.Tensor`):
            A 1D tensor of size equal to the number of classes. Specifies the manual
            weight rescaling applied to each class. Useful in cases when there is severe
            class imbalance in the training set.
        num_classes (int):
            The number of classes.
        size_average (bool):
            By default, the losses are averaged for each minibatch over observations **as
            well as** over dimensions. However, if ``False`` the losses are instead
            summed. This is a keyword only parameter.
    """

    def __init__(self, label_smoothing=0, weight=None, num_classes=2, **kwargs):
        super(SoftNLLLoss, self).__init__(**kwargs)
        self.label_smoothing = label_smoothing
        self.confidence = 1 - self.label_smoothing
        self.num_classes = num_classes
        self.register_buffer('weight', Variable(weight))

        assert label_smoothing >= 0.0 and label_smoothing <= 1.0

        self.criterion = nn.KLDivLoss(**kwargs)

    def forward(self, input, target):
        one_hot = torch.zeros_like(input)
        one_hot.fill_(self.label_smoothing / (self.num_classes - 1))
        one_hot.scatter_(1, target.unsqueeze(1).long(), self.confidence)

        if self.weight is not None:
            one_hot.mul_(self.weight)

        return self.criterion(input, one_hot)


# This class is based on the Optimizer class in the ONMT-py project.
class Optimizer(object):
    """Controller class for optimization.

    Mostly a thin wrapper for `optim`, but also useful for implementing learning rate
    scheduling beyond what is currently available. Also implements necessary methods for
    training RNNs such as grad manipulations.

    Args:
        method (string):
            One of [sgd, adagrad, adadelta, adam].
        lr (float):
            Learning rate.
        lr_decay (float):
            Learning rate decay multiplier.
        start_decay_at (int):
            Epoch to start learning rate decay. If None, starts decay when the validation
            accuracy stops improving. Defaults to 1.
        beta1, beta2 (float):
            Hyperarameters for adam.
        adagrad_accum (float, optional):
            Initialization hyperparameter for adagrad.
    """

    def __init__(self,
                 method='adam',
                 lr=0.001,
                 max_grad_norm=5,
                 start_decay_at=1,
                 beta1=0.9,
                 beta2=0.999,
                 adagrad_accum=0.0,
                 lr_decay=0.8):
        self.last_acc = None
        self.lr = lr
        self.original_lr = lr
        self.max_grad_norm = max_grad_norm
        self.method = method
        self.lr_decay = lr_decay
        self.start_decay_at = start_decay_at
        self.start_decay = False
        self._step = 0
        self.betas = [beta1, beta2]
        self.adagrad_accum = adagrad_accum
        self.params = None
        logger.info('Initial learning rate: {:0.3e}'.format(self.lr))

    def set_parameters(self, params):
        """Sets the model parameters and initializes the base optimizer.

        Args:
            params:
                Dictionary of named model parameters. Parameters that do not require
                gradients will be filtered out for optimization.
        """
        self.params = []
        for k, p in params:
            if p.requires_grad:
                self.params.append(p)
        if self.method == 'sgd':
            self.base_optimizer = optim.SGD(self.params, lr=self.lr)
        elif self.method == 'adagrad':
            self.base_optimizer = optim.Adagrad(self.params, lr=self.lr)
            for group in self.base_optimizer.param_groups:
                for p in group['params']:
                    self.base_optimizer.state[p]['sum'] = self.base_optimizer\
                        .state[p]['sum'].fill_(self.adagrad_accum)
        elif self.method == 'adadelta':
            self.base_optimizer = optim.Adadelta(self.params, lr=self.lr)
        elif self.method == 'adam':
            self.base_optimizer = optim.Adam(
                self.params, lr=self.lr, betas=self.betas, eps=1e-9)
        else:
            raise RuntimeError("Invalid optim method: " + self.method)

    def _set_rate(self, lr):
        for param_group in self.base_optimizer.param_groups:
            param_group['lr'] = self.lr

    def step(self):
        """Update the model parameters based on current gradients.

        Optionally, will employ gradient clipping.
        """
        self._step += 1

        if self.max_grad_norm:
            clip_grad_norm_(self.params, self.max_grad_norm)
        self.base_optimizer.step()

    def update_learning_rate(self, acc, epoch):
        """Decay learning rate.

        Decays lerning rate if val perf does not improve or we hit the `start_decay_at`
        limit.

        Args:
            acc:
                The accuracy score on the validation set.
            epoch:
                The current epoch number.
        """

        if self.start_decay_at is not None and epoch >= self.start_decay_at:
            self.start_decay = True
        if self.last_acc is not None and acc < self.last_acc:
            self.start_decay = True

        if self.start_decay:
            self.lr = self.lr * self.lr_decay
            logger.info('Setting learning rate to {:0.3e} for next epoch'.format(self.lr))

        self.last_acc = acc
        self._set_rate(self.lr)



from .field import MatchingField, reset_vector_cache
from .dataset import MatchingDataset
from .iterator import MatchingIterator
from .process import process, process_unlabeled
from .dataset import split

__all__ = [
    'MatchingField', 'MatchingDataset', 'MatchingIterator', 'process', 'process_unlabeled', 'split',
    'reset_vector_cache'
]



import copy
import logging
from collections import Mapping

import dill
import six

import deepmatcher as dm
import torch
import torch.nn as nn

from . import _utils
from ..data import MatchingDataset, MatchingIterator
from ..runner import Runner
from ..utils import Bunch, tally_parameters

logger = logging.getLogger('deepmatcher.core')


class MatchingModel(nn.Module):
    r"""A neural network model for entity matching.

    Refer to the
    `Matching Models tutorial <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb>`_
    for details on how to customize a `MatchingModel`. A brief intro is below:

    This network consists of the following components:

    #. Attribute Summarizers

    #. Attribute Comparators

    #. A Classifier

    Creating a MatchingModel instance does not immediately construct the neural network.
    The network will be constructed just before training based on metadata from the
    training set:

    #. For each attribute (e.g., Product Name, Address, etc.), an Attribute Summarizer is
       constructed using the specified `attr_summarizer` template.
    #. For each attribute, an Attribute Comparator is constructed using the specified
       `attr_summarizer` template.
    #. A Classifier is constructed based on the specified `classifier` template.

    Args:
        attr_summarizer (string or :class:`AttrSummarizer` or callable):
            The attribute summarizer. Takes in two word embedding sequences and summarizes
            the information in them to produce two summary vectors as output.
            Options `listed here <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#attr_summarizer-can-be-set-to-one-of-the-following:>`__.
            Defaults to 'hybrid', i.e., the :class:`~deepmatcher.attr_summarizers.Hybrid`
            model.
        attr_comparator (string or :class:`~deepmatcher.modules.Merge` or callable):
            The attribute comparator. Takes as input the two summary vectors and applies a
            comparison function over those summaries to obtain the final similarity
            representation of the two attribute values. Argument must specify a
            :ref:`merge-op` operation. Default is selected based on `attr_summarizer`
            choice.
        attr_condense_factor (string or int):
            The factor by which to condense each attribute similarity vector. E.g. if
            `attr_condense_factor` is set to 3 and the attribute similarity vector size is
            300, then each attribute similarity vector is transformed to a 100 dimensional
            vector using a linear transformation. The purpose of condensing is to reduce
            the number of parameters in the classifier module. This parameter can be set
            to a number or 'auto'. If 'auto', then the condensing factor is set to be
            equal to the number attributes, but if there are more than 6 attributes, then
            the condensing factor is set to 6. Defaults to 'auto'.
        attr_merge (string or :class:`~deepmatcher.modules.Merge` or callable):
            The operation used to merge the (optionally condensed) attribute similarity
            vectors to obtain the input to the classifier. Argument must specify a
            :ref:`merge-op` operation. Defaults to 'concat', i.e., concatenate all
            attribute similarity vectors to form the classifier input.
        classifier (string or :class:`Classifier` or callable):
            The neural network to perform match / mismatch classification
            based on attribute similarity representations.
            Options `listed here <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#classifier-can-be-set-to-one-of-the-following:>`__.
            Defaults to '2-layer-highway', i.e., use a two layer highway network followed
            by a softmax layer for classification.
        hidden_size (int):
            The hidden size to use for the `attr_summarizer` and the `classifier` modules,
            if they are string arguments. If a module or :attr:`callable` input is specified
            for a component, this argument is ignored for that component.
    """

    def __init__(self,
                 attr_summarizer='hybrid',
                 attr_condense_factor='auto',
                 attr_comparator=None,
                 attr_merge='concat',
                 classifier='2-layer-highway',
                 hidden_size=300):

        super(MatchingModel, self).__init__()

        self.attr_summarizer = attr_summarizer
        self.attr_condense_factor = attr_condense_factor
        self.attr_comparator = attr_comparator
        self.attr_merge = attr_merge
        self.classifier = classifier

        self.hidden_size = hidden_size
        self._train_buffers = set()
        self._initialized = False

    def run_train(self, *args, **kwargs):
        """run_train(train_dataset, validation_dataset, best_save_path,epochs=30, \
            criterion=None, optimizer=None, pos_neg_ratio=None, pos_weight=None, \
            label_smoothing=0.05, save_every_prefix=None, save_every_freq=None, \
            batch_size=32, device=None, progress_style='bar', log_freq=5, \
            sort_in_buckets=None)

        Train the model using the specified training set.

        Args:
            train_dataset (:class:`~deepmatcher.data.MatchingDataset`):
                The training dataset obtained using :func:`deepmatcher.data.process`.
            validation_dataset (:class:`~deepmatcher.data.MatchingDataset`):
                The validation dataset obtained using :func:`deepmatcher.data.process`.
                This is used for `early stopping
                <https://en.wikipedia.org/wiki/Early_stopping>`_.
            best_save_path (string):
                The path to save the best model to. At the end of each epoch, if the new
                model accuracy (F1) is better than all previous epochs, then it is saved
                to this location.
            epochs (int):
                Number of training epochs, i.e., number of times to cycle through the
                entire training set. Defaults to 50.
            criterion (:class:`torch.nn.Module`):
                The loss function to use. Refer to the `losses section of the PyTorch API
                <https://pytorch.org/docs/master/nn.html#loss-functions>`_ for options. By
                default, `deepmatcher` will output a 2d tensor of shape (N, C) where N is
                the batch size and C is 2 - the number of classes. Keep this in mind when
                picking the loss. Defaults to :class:`~deepmatcher.optim.SoftNLLLoss` with
                label smoothing.
            optimizer (:class:`~deepmatcher.optim.Optimizer`):
                The optimizer to use for updating the trainable parameters of the
                :class:`~deepmatcher.MatchingModel` neural network after each iteration.
                If not specified an :class:`~deepmatcher.optim.Optimizer` with Adam
                optimizer will be constructed.
            pos_neg_ratio (int):
                The weight of the positive class (match) wrt the negative class
                (non-match). This parameter must be specified if there is a significant
                class imbalance in the dataset.
            label_smoothing (float):
                The `label_smoothing` parameter to constructor of
                :class:`~deepmatcher.optim.SoftNLLLoss` criterion. Only used when
                `criterion` param is None. Defaults to 0.05.
            save_every_prefix (string):
                Prefix of the path to save model to, after end of
                every N epochs, where N is determined by `save_every_freq` param.
                E.g. setting this to "/path/to/saved/model" will save models to
                "/path/to/saved/model_ep1.pth", "/path/to/saved/model_ep2.pth", etc.
                Models will not be saved periodically if this is None. Defaults to
                None.
            save_every_freq (int):
                Determines the frequency (number of epochs) for saving models periodically
                to the path specified by the `save_every_prefix` param (has no effect if
                that param is not set). Defaults to 1.
            batch_size (int):
                Mini-batch size for SGD. For details on what this is
                `see this video <https://www.coursera.org/learn/machine-learning/lecture/9zJUs/mini-batch-gradient-descent>`__.
                Defaults to 32. This is a keyword only param.
            device (string):
                The device on which to train the model ('cpu' or 'cuda'). If None, will use
                first available GPU, or use CPU if no GPUs are available. Defaults to None.
                This is a keyword only param.
            progress_style (string):
                Sets the progress update style. One of 'bar' or 'log'. If 'bar', uses a
                progress bar, updated every N batches. If 'log', prints training stats
                every N batches. N is determined by the `log_freq` param.
                This is a keyword only param.
            log_freq (int):
                Number of batches between progress updates. Defaults to 5.
                This is a keyword only param.
            sort_in_buckets (bool):
                Whether to batch examples of similar lengths together. If True, minimizes
                amount of padding needed while producing freshly shuffled batches for each
                new epoch. Implemented using :func:`torchtext.data.pool`.
                Defaults to True. This is a keyword only param.

        Returns:
            float: The best F1 score obtained by the model on the validation dataset.
        """
        return Runner.train(self, *args, **kwargs)

    def run_eval(self, *args, **kwargs):
        """run_eval(dataset, batch_size=32, device=None, progress_style='bar', \
            log_freq=5, sort_in_buckets=None)

        Evaluate the model on the specified dataset.

        Args:
            dataset (:class:`~deepmatcher.data.MatchingDataset`): The evaluation dataset
                obtained using :func:`deepmatcher.data.process`.
            batch_size (int):
                Mini-batch size for SGD. For details on what this is
                `see this video <https://www.coursera.org/learn/machine-learning/lecture/9zJUs/mini-batch-gradient-descent>`__.
                Defaults to 32. This is a keyword only param.
            device (string):
                The device on which to train the model ('cpu' or 'cuda'). If None, will use
                first available GPU, or use CPU if no GPUs are available. Defaults to None.
                This is a keyword only param.
            progress_style (string):
                Sets the progress update style. One of 'bar' or 'log'. If 'bar', uses a
                progress bar, updated every N batches. If 'log', prints training stats
                every N batches. N is determined by the `log_freq` param.
                This is a keyword only param.
            log_freq (int):
                Number of batches between progress updates. Defaults to 5.
                This is a keyword only param.
            sort_in_buckets (bool):
                Whether to batch examples of similar lengths together. If True, minimizes
                amount of padding needed while producing freshly shuffled batches for each
                new epoch. Implemented using :func:`torchtext.data.pool`.
                Defaults to True. This is a keyword only param.

        Returns:
            float: The F1 score obtained by the model on the dataset.
        """
        return Runner.eval(self, *args, **kwargs)

    def run_prediction(self, *args, **kwargs):
        """run_prediction(dataset, output_attributes=False, batch_size=32, device=None, \
            progress_style='bar', log_freq=5, sort_in_buckets=None)

        Use the model to obtain predictions, i.e., match scores on the specified dataset.

        Args:
            dataset (:class:`~deepmatcher.data.MatchingDataset`): The dataset (labeled or
                not) obtained using :func:`deepmatcher.data.process` or
                :func:`deepmatcher.data.process_unlabeled`.
            output_attributes (bool): Whether to include all attributes in the original
                CSV file of the dataset in the returned pandas table.
            batch_size (int):
                Mini-batch size for SGD. For details on what this is
                `see this video <https://www.coursera.org/learn/machine-learning/lecture/9zJUs/mini-batch-gradient-descent>`__.
                Defaults to 32. This is a keyword only param.
            device (string):
                The device on which to train the model ('cpu' or 'cuda'). If None, will use
                first available GPU, or use CPU if no GPUs are available. Defaults to None.
                This is a keyword only param.
            progress_style (string):
                Sets the progress update style. One of 'bar' or 'log'. If 'bar', uses a
                progress bar, updated every N batches. If 'log', prints training stats
                every N batches. N is determined by the `log_freq` param.
                This is a keyword only param.
            log_freq (int):
                Number of batches between progress updates. Defaults to 5.
                This is a keyword only param.
            sort_in_buckets (bool):
                Whether to batch examples of similar lengths together. If True, minimizes
                amount of padding needed while producing freshly shuffled batches for each
                new epoch. Implemented using :func:`torchtext.data.pool`.
                Defaults to True. This is a keyword only param.


        Returns:
            pandas.DataFrame: A pandas DataFrame containing tuple pair IDs (in the "id"
                column) and the corresponding match score predictions (in the
                "match_score" column). Will also include all attributes in the original
                CSV file of the dataset if `output_attributes` is True.
        """
        return Runner.predict(self, *args, **kwargs)

    def initialize(self, train_dataset, init_batch=None):
        r"""Initialize (not lazily) the matching model given the actual training data.

        Instantiates all sub-components and their trainable parameters.

        Args:
            train_dataset (:class:`~deepmatcher.data.MatchingDataset`):
                The training dataset obtained using :func:`deepmatcher.data.process`.
            init_batch (:class:`~deepmatcher.batch.MatchingBatch`):
                A batch of data to forward propagate through the model. If None, a batch
                is drawn from the training dataset.
        """

        if self._initialized:
            return

        # Copy over training info from train set for persistent state. But remove actual
        # data examples.
        self.meta = Bunch(**train_dataset.__dict__)
        if hasattr(self.meta, 'fields'):
            del self.meta.fields
            del self.meta.examples

        self._register_train_buffer('state_meta', Bunch(**self.meta.__dict__))
        del self.state_meta.metadata  # we only need `self.meta.orig_metadata` for state.

        self.attr_summarizers = dm.modules.ModuleMap()
        if isinstance(self.attr_summarizer, Mapping):
            for name, summarizer in self.attr_summarizer.items():
                self.attr_summarizers[name] = AttrSummarizer._create(
                    summarizer, hidden_size=self.hidden_size)
            assert len(
                set(self.attr_summarizers.keys()) ^ set(self.meta.canonical_text_fields)
            ) == 0
        else:
            self.attr_summarizer = AttrSummarizer._create(
                self.attr_summarizer, hidden_size=self.hidden_size)
            for name in self.meta.canonical_text_fields:
                self.attr_summarizers[name] = copy.deepcopy(self.attr_summarizer)

        if self.attr_condense_factor == 'auto':
            self.attr_condense_factor = min(len(self.meta.canonical_text_fields), 6)
            if self.attr_condense_factor == 1:
                self.attr_condense_factor = None

        if not self.attr_condense_factor:
            self.attr_condensors = None
        else:
            self.attr_condensors = dm.modules.ModuleMap()
            for name in self.meta.canonical_text_fields:
                self.attr_condensors[name] = dm.modules.Transform(
                    '1-layer-highway',
                    non_linearity=None,
                    output_size=self.hidden_size // self.attr_condense_factor)

        self.attr_comparators = dm.modules.ModuleMap()
        if isinstance(self.attr_comparator, Mapping):
            for name, comparator in self.attr_comparator.items():
                self.attr_comparators[name] = _create_attr_comparator(comparator)
            assert len(
                set(self.attr_comparators.keys()) ^ set(self.meta.canonical_text_fields)
            ) == 0
        else:
            if isinstance(self.attr_summarizer, AttrSummarizer):
                self.attr_comparator = self._get_attr_comparator(
                    self.attr_comparator, self.attr_summarizer)
            else:
                if self.attr_comparator is None:
                    raise ValueError('"attr_comparator" must be specified if '
                                     '"attr_summarizer" is custom.')

            self.attr_comparator = _create_attr_comparator(self.attr_comparator)
            for name in self.meta.canonical_text_fields:
                self.attr_comparators[name] = copy.deepcopy(self.attr_comparator)

        self.attr_merge = dm.modules._merge_module(self.attr_merge)
        self.classifier = _utils.get_module(
            Classifier, self.classifier, hidden_size=self.hidden_size)

        self._reset_embeddings(train_dataset.vocabs)

        # Instantiate all components using a small batch from training set.
        if not init_batch:
            run_iter = MatchingIterator(
                train_dataset,
                train_dataset,
                train=False,
                batch_size=4,
                device='cpu',
                sort_in_buckets=False)
            init_batch = next(run_iter.__iter__())
        self.forward(init_batch)

        # Keep this init_batch for future initializations.
        self.state_meta.init_batch = init_batch

        self._initialized = True
        logger.info('Successfully initialized MatchingModel with {:d} trainable '
                    'parameters.'.format(tally_parameters(self)))

    def _reset_embeddings(self, vocabs):
        self.embed = dm.modules.ModuleMap()
        field_vectors = {}
        for name in self.meta.all_text_fields:
            vectors = vocabs[name].vectors
            if vectors not in field_vectors:
                vectors_size = vectors.shape
                embed = nn.Embedding(vectors_size[0], vectors_size[1])
                embed.weight.data.copy_(vectors)
                embed.weight.requires_grad = False
                field_vectors[vectors] = dm.modules.NoMeta(embed)
            self.embed[name] = field_vectors[vectors]

    def _get_attr_comparator(self, arg, attr_summarizer):
        r"""Get the attribute comparator.

        Args:
            arg (string):
                The attribute comparator to use. Can be one of the supported style
                arguments for :class:`~modules.Merge`, specifically, 'abs-diff', 'diff',
                'concat', 'concat-diff', 'concat-abs-diff', or 'mul'. If not specified
                uses 'abs-diff' for SIF and RNN, 'concat' for Attention, and 'concat-diff'
                for Hybrid".
            attr_summarizer (:class:`deepmatcher.AttrSummarizer`):
                The attribute summarizer object in :mod:`attr_summarizers` (i.e., it
                should be one of "SIF", "RNN", "Attention", and "Hybrid" defined in
                :mod:`attr_summarizers`).

        Returns:
            string: Return the type of attribute comparator.
        """
        if arg is not None:
            return arg
        if isinstance(attr_summarizer, dm.attr_summarizers.SIF):
            return 'abs-diff'
        elif isinstance(attr_summarizer, dm.attr_summarizers.RNN):
            return 'abs-diff'
        elif isinstance(attr_summarizer, dm.attr_summarizers.Attention):
            return 'concat'
        elif isinstance(attr_summarizer, dm.attr_summarizers.Hybrid):
            return 'concat-abs-diff'
        raise ValueError('Cannot infer attr comparator, please specify.')

    def forward(self, input):
        r"""Performs a forward pass through the model.

        Overrides :meth:`torch.nn.Module.forward`.

        Args:
            input (:class:`~deepmatcher.batch.MatchingBatch`): A batch of tuple pairs
                processed into tensors.
        """
        embeddings = {}
        for name in self.meta.all_text_fields:
            attr_input = getattr(input, name)
            embeddings[name] = self.embed[name](attr_input)

        attr_comparisons = []
        for name in self.meta.canonical_text_fields:
            left, right = self.meta.text_fields[name]
            left_summary, right_summary = self.attr_summarizers[name](embeddings[left],
                                                                      embeddings[right])

            # Remove metadata information at this point.
            left_summary, right_summary = left_summary.data, right_summary.data

            if self.attr_condensors:
                left_summary = self.attr_condensors[name](left_summary)
                right_summary = self.attr_condensors[name](right_summary)
            attr_comparisons.append(self.attr_comparators[name](left_summary,
                                                                right_summary))

        entity_comparison = self.attr_merge(*attr_comparisons)
        return self.classifier(entity_comparison)

    def _register_train_buffer(self, name, value):
        r"""Adds a persistent buffer containing training metadata to the module.
        """
        self._train_buffers.add(name)
        setattr(self, name, value)

    def save_state(self, path, include_meta=True):
        r"""Save the model state to a certain path.

        Args:
            path (string): The path to save the model state to.
            include_meta (bool): Whether to include training dataset metadata along with
                the model parameters when saving. If False, the model will not be
                automatically initialized upon loading - you will need to initialize
                manually using :meth:`initialize`.
        """
        state = {'model': self.state_dict()}
        for k in self._train_buffers:
            if include_meta or k != 'state_meta':
                state[k] = getattr(self, k)
        torch.save(state, path, pickle_module=dill)

    def load_state(self, path, map_location=None):
        r"""Load the model state from a file in a certain path.

        Args:
            path (string): The path to load the model state from.
        """
        state = torch.load(path, pickle_module=dill, map_location=map_location)
        for k, v in six.iteritems(state):
            if k != 'model':
                self._train_buffers.add(k)
                setattr(self, k, v)

        if hasattr(self, 'state_meta'):
            train_info = copy.copy(self.state_meta)

            # Handle metadata manually.
            # TODO (Sid): Make this cleaner.
            train_info.metadata = train_info.orig_metadata
            MatchingDataset.finalize_metadata(train_info)

            self.initialize(train_info, self.state_meta.init_batch)

        self.load_state_dict(state['model'])


class AttrSummarizer(dm.modules.LazyModule):
    r"""__init__(word_contextualizer, word_comparator, word_aggregator, hidden_size=None)

    The Attribute Summarizer.

    Summarizes the two word embedding sequences of an attribute.
    `Refer this tutorial <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#2.1.-Attribute-Summarization>`__
    for details. Sub-classes that implement various built-in options for this module are
    defined in :mod:`deepmatcher.attr_summarizers`.

    Args:
        word_contextualizer (string or :class:`WordContextualizer` or callable):
            Module that takes a word embedding sequence and produces a context-aware
            word embedding sequence as output.
            Options `listed here <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#word_contextualizer-can-be-set-to-one-of-the-following:>`__.
        word_comparator (string or :class:`WordComparator` or callable):
            Module that takes two word embedding sequences, aligns words in the two
            sequences, and performs a word by word comparison.
            Options `listed here <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#word_comparator-can-be-set-to-one-of-the-following:>`__.
        word_aggregator (string or :class:`WordAggregator` or callable):
            Module that takes a sequence of vectors and aggregates it into a single
            vector.
            Options `listed here <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#word_aggregator-can-be-set-to-one-of-the-following:>`__.
        hidden_size (int):
            The hidden size used for the three sub-modules (i.e., word_contextualizer,
            word_comparator, and word_aggregator). If None, uses the input size, i.e., the
            size of the last dimension of the input to this module as the hidden size.
            Defaults to None.
    """

    def _init(self,
              word_contextualizer,
              word_comparator,
              word_aggregator,
              hidden_size=None):
        self.word_contextualizer = WordContextualizer._create(
            word_contextualizer, hidden_size=hidden_size)
        self.word_comparator = WordComparator._create(
            word_comparator, hidden_size=hidden_size)
        self.word_aggregator = WordAggregator._create(
            word_aggregator, hidden_size=hidden_size)

    def _forward(self, left_input, right_input):
        r"""The forward function of attribute summarizer.
        """
        left_contextualized, right_contextualized = left_input, right_input
        if self.word_contextualizer:
            left_contextualized = self.word_contextualizer(left_input)
            right_contextualized = self.word_contextualizer(right_input)

        left_compared, right_compared = left_contextualized, right_contextualized
        if self.word_comparator:
            left_compared = self.word_comparator(
                left_contextualized, right_contextualized, left_input, right_input)
            right_compared = self.word_comparator(
                right_contextualized, left_contextualized, right_input, left_input)

        left_aggregator_context = right_input
        right_aggregator_context = left_input

        left_aggregated = self.word_aggregator(left_compared, left_aggregator_context)
        right_aggregated = self.word_aggregator(right_compared, right_aggregator_context)
        return left_aggregated, right_aggregated

    @classmethod
    def _create(cls, arg, **kwargs):
        r"""Create an attribute summarization object.

        Args:
            arg (string or :mod:`deepmatcher.attr_summarizers` or callable):
                Same as the `attr_summarizer` argument to the constructor of
                :class:`MatchingModel`.
            **kwargs:
                Keyword arguments to the constructor of the AttrSummarizer sub-class.
                For details on what these can be, please refer to the documentation of the
                sub-classes in :mod:`deepmatcher.attr_summarizers`.
        """
        assert arg is not None
        if isinstance(arg, six.string_types):
            type_map = {
                'sif': dm.attr_summarizers.SIF,
                'rnn': dm.attr_summarizers.RNN,
                'attention': dm.attr_summarizers.Attention,
                'hybrid': dm.attr_summarizers.Hybrid
            }
            if arg in type_map:
                asr = type_map[arg](**kwargs)
            else:
                raise ValueError('Unknown Attribute Summarizer name.')
        else:
            asr = _utils.get_module(AttrSummarizer, arg)

        asr.expect_signature('[AxBxC, AxDxC] -> [AxE, AxE]')
        return asr


def _create_attr_comparator(arg):
    r"""Create an attribute comparator object.

    Args:
        arg (string):
            Same as the `attr_comparator` argument to the constructor of
            :class:`MatchingModel`.
    """
    assert arg is not None
    module = dm.modules._merge_module(arg)
    module.expect_signature('[AxB, AxB] -> [AxC]')
    return module


class WordContextualizer(dm.modules.LazyModule):
    r"""__init__()

    The Word Contextualizer.

    Takes a word embedding sequence and produces a context-aware word embedding sequence.
    `Refer this tutorial <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#2.1.1.-Word-Contextualizer>`__
    for details. Sub-classes that implement various options for this module are defined
    in :mod:`deepmatcher.word_contextualizers`.
    """

    @classmethod
    def _create(cls, arg, **kwargs):
        r"""Create a word contextualizer object.

        Args:
            arg (string or :mod:`deepmatcher.word_contextualizers` or callable):
                Same as the `word_contextualizer` argument to the constructor of
                :class:`AttrSummarizer`.
            **kwargs:
                Keyword arguments to the constructor of the WordContextualizer sub-class.
                For details on what these can be, please refer to the documentation of the
                sub-classes in :mod:`deepmatcher.word_contextualizers`.
        """
        if isinstance(arg, six.string_types):
            if dm.word_contextualizers.RNN.supports_style(arg):
                wc = dm.word_contextualizers.RNN(arg, **kwargs)
            elif arg == 'self-attention':
                wc = dm.word_contextualizers.SelfAttention(**kwargs)
            else:
                raise ValueError('Unknown Word Contextualizer name.')
        else:
            wc = _utils.get_module(WordContextualizer, arg)

        if wc is not None:
            wc.expect_signature('[AxBxC] -> [AxBxD]')

        return wc


class WordComparator(dm.modules.LazyModule):
    r"""__init__()

    The Word Comparator.

    Takes two word embedding sequences, aligns words in the two sequences, and performs a
    word by word comparison.
    `Refer this tutorial <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#2.1.2.-Word-Comparator>`__
    for details. Sub-classes that implement various options for this module are defined
    in :mod:`deepmatcher.word_comparators`.
    """

    @classmethod
    def _create(cls, arg, **kwargs):
        r"""Create a word comparator object.

        Args:
            arg (string or :mod:`deepmatcher.word_comparators` or callable):
                Same as the `word_comparator` argument to the constructor of
                :class:`AttrSummarizer`.
            **kwargs:
                Keyword arguments to the constructor of the WordComparator sub-class.
                For details on what these can be, please refer to the documentation of the
                sub-classes in :mod:`deepmatcher.word_comparators`.
        """
        if isinstance(arg, six.string_types):
            parts = arg.split('-')
            if (parts[1] == 'attention' and
                    dm.modules.AlignmentNetwork.supports_style(parts[0])):
                wc = dm.word_comparators.Attention(alignment_network=parts[0], **kwargs)
            else:
                raise ValueError('Unknown Word Comparator name.')
        else:
            wc = _utils.get_module(WordComparator, arg)

        if wc is not None:
            wc.expect_signature('[AxBxC, AxDxC, AxBxE, AxDxE] -> [AxBxF]')

        return wc


class WordAggregator(dm.modules.LazyModule):
    r"""__init__()

    The Word Aggregator.

    Takes a sequence of vectors and aggregates it into a single vector.
    `Refer this tutorial <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#2.1.3.-Word-Aggregator>`__
    for details. Sub-classes that implement various options for this module are defined in
    :mod:`deepmatcher.word_aggregators`.
    """

    @classmethod
    def _create(cls, arg, **kwargs):
        r"""
        Create a word aggregator object.

        Args:
            arg (string or :mod:`deepmatcher.word_aggregators` or callable):
                Same as the `word_aggregator` argument to the constructor of
                :class:`AttrSummarizer`.
        **kwargs:
            Keyword arguments to the constructor of the WordAggregator sub-class.
            For details on what these can be, please refer to the documentation of the
            sub-classes in :mod:`deepmatcher.word_aggregators`.
        """
        assert arg is not None
        if isinstance(arg, six.string_types):
            parts = arg.split('-')
            if (parts[-1] == 'pool' and
                    dm.word_aggregators.Pool.supports_style('-'.join(parts[:-1]))):
                seq = []
                seq.append(dm.modules.Lambda(lambda x1, x2: x1))  # Ignore the context.
                seq.append(dm.word_aggregators.Pool(style='-'.join(parts[:-1])))

                # Make lazy module.
                wa = dm.modules.LazyModuleFn(lambda: dm.modules.MultiSequential(*seq))
            elif arg == 'attention-with-rnn':
                wa = dm.word_aggregators.AttentionWithRNN(**kwargs)
            else:
                raise ValueError('Unknown Word Aggregator name.')
        else:
            wa = _utils.get_module(WordAggregator, arg)

        wa.expect_signature('[AxBxC] -> [AxD]')
        return wa


class Classifier(nn.Sequential):
    r"""The Classifier Network.

    Predicts whether a tuple pair matches or not given a representation of all the
    attribute summarizations.
    `Refer this tutorial <https://nbviewer.jupyter.org/github/sidharthms/deepmatcher/blob/master/examples/matching_models.ipynb#3.-Classifier>`__
    for details.

    Args:
        transform_network (string or :class:`~deepmatcher.modules.Transform` or callable):
            The neural network to transform the input vector of the classifier to a
            hidden representation of size `hidden_size`. Argument must specify a
            :ref:`transform-op` operation.

        hidden_size (int):
            The size of the hidden representation generated by the transformation network.
            If None, uses the size of the input vector to this module as the hidden size.
    """

    def __init__(self, transform_network, hidden_size=None):
        super(Classifier, self).__init__()
        if transform_network:
            self.add_module('transform',
                            dm.modules._transform_module(transform_network, hidden_size))
        self.add_module('softmax_transform',
                        dm.modules.Transform(
                            '1-layer', non_linearity=None, output_size=2))
        self.add_module('softmax', nn.LogSoftmax(dim=1))



import copy
import logging
import sys
import time
import warnings
from collections import OrderedDict

import pandas as pd

import pyprind
import torch
from tqdm import tqdm

from .data import MatchingIterator
from .optim import Optimizer, SoftNLLLoss
from .utils import tally_parameters

try:
    get_ipython
    from tqdm import tqdm_notebook as tqdm
except NameError:
    from tqdm import tqdm

logger = logging.getLogger(__name__)


class Statistics(object):
    """Accumulator for loss statistics, inspired by ONMT.

    Keeps track of the following metrics:
    * F1
    * Precision
    * Recall
    * Accuracy
    """

    def __init__(self):
        self.loss_sum = 0
        self.examples = 0
        self.tps = 0
        self.tns = 0
        self.fps = 0
        self.fns = 0
        self.start_time = time.time()

    def update(self, loss=0, tps=0, tns=0, fps=0, fns=0):
        examples = tps + tns + fps + fns
        self.loss_sum += loss * examples
        self.tps += tps
        self.tns += tns
        self.fps += fps
        self.fns += fns
        self.examples += examples

    def loss(self):
        return self.loss_sum / self.examples

    def f1(self):
        prec = self.precision()
        recall = self.recall()
        return 2 * prec * recall / max(prec + recall, 1)

    def precision(self):
        return 100 * self.tps / max(self.tps + self.fps, 1)

    def recall(self):
        return 100 * self.tps / max(self.tps + self.fns, 1)

    def accuracy(self):
        return 100 * (self.tps + self.tns) / self.examples

    def examples_per_sec(self):
        return self.examples / (time.time() - self.start_time)


class Runner(object):
    """Experiment runner.

    This class implements routines to train, evaluate and make predictions from models.
    """

    @staticmethod
    def _print_stats(name, epoch, batch, n_batches, stats, cum_stats):
        """Write out batch statistics to stdout.
        """
        print((' | {name} | [{epoch}][{batch:4d}/{n_batches}] || Loss: {loss:7.4f} |'
               ' F1: {f1:7.2f} | Prec: {prec:7.2f} | Rec: {rec:7.2f} ||'
               ' Cum. F1: {cf1:7.2f} | Cum. Prec: {cprec:7.2f} | Cum. Rec: {crec:7.2f} ||'
               ' Ex/s: {eps:6.1f}').format(
                   name=name,
                   epoch=epoch,
                   batch=batch,
                   n_batches=n_batches,
                   loss=stats.loss(),
                   f1=stats.f1(),
                   prec=stats.precision(),
                   rec=stats.recall(),
                   cf1=cum_stats.f1(),
                   cprec=cum_stats.precision(),
                   crec=cum_stats.recall(),
                   eps=cum_stats.examples_per_sec()))

    @staticmethod
    def _print_final_stats(epoch, runtime, datatime, stats):
        """Write out epoch statistics to stdout.
        """
        print(('Finished Epoch {epoch} || Run Time: {runtime:6.1f} | '
               'Load Time: {datatime:6.1f} || F1: {f1:6.2f} | Prec: {prec:6.2f} | '
               'Rec: {rec:6.2f} || Ex/s: {eps:6.2f}\n').format(
                   epoch=epoch,
                   runtime=runtime,
                   datatime=datatime,
                   f1=stats.f1(),
                   prec=stats.precision(),
                   rec=stats.recall(),
                   eps=stats.examples_per_sec()))

    @staticmethod
    def _set_pbar_status(pbar, stats, cum_stats):
        postfix_dict = OrderedDict([
            ('Loss', '{0:7.4f}'.format(stats.loss())),
            ('F1', '{0:7.2f}'.format(stats.f1())),
            ('Cum. F1', '{0:7.2f}'.format(cum_stats.f1())),
            ('Ex/s', '{0:6.1f}'.format(cum_stats.examples_per_sec())),
        ])
        pbar.set_postfix(ordered_dict=postfix_dict)

    @staticmethod
    def _compute_scores(output, target):
        predictions = output.max(1)[1].data
        correct = (predictions == target.data).float()
        incorrect = (1 - correct).float()
        positives = (target.data == 1).float()
        negatives = (target.data == 0).float()

        tp = torch.dot(correct, positives)
        tn = torch.dot(correct, negatives)
        fp = torch.dot(incorrect, negatives)
        fn = torch.dot(incorrect, positives)

        return tp, tn, fp, fn

    @staticmethod
    def _run(run_type,
             model,
             dataset,
             criterion=None,
             optimizer=None,
             train=False,
             device=None,
             batch_size=32,
             batch_callback=None,
             epoch_callback=None,
             progress_style='bar',
             log_freq=5,
             sort_in_buckets=None,
             return_predictions=False,
             **kwargs):

        if device is None:
            device = 'cuda' if torch.cuda.is_available() else 'cpu'
        elif device == 'gpu':
            device = 'cuda'

        sort_in_buckets = train
        run_iter = MatchingIterator(
            dataset,
            model.meta,
            train,
            batch_size=batch_size,
            device=device,
            sort_in_buckets=sort_in_buckets)

        model = model.to(device)
        if criterion:
            criterion = criterion.to(device)

        if train:
            model.train()
        else:
            model.eval()

        epoch = model.epoch
        datatime = 0
        runtime = 0
        cum_stats = Statistics()
        stats = Statistics()
        predictions = []
        id_attr = model.meta.id_field
        label_attr = model.meta.label_field

        if train and epoch == 0:
            print('* Number of trainable parameters:', tally_parameters(model))

        epoch_str = 'Epoch {0:d}'.format(epoch + 1)
        print('===> ', run_type, epoch_str)
        batch_end = time.time()

        # The tqdm-bar for Jupyter notebook is under development.
        if progress_style == 'tqdm-bar':
            pbar = tqdm(
                total=len(run_iter) // log_freq,
                bar_format='{l_bar}{bar}{postfix}',
                file=sys.stdout)

        # Use the pyprind bar as the default progress bar.
        if progress_style == 'bar':
            pbar = pyprind.ProgBar(len(run_iter) // log_freq, bar_char='█', width=30)

        for batch_idx, batch in enumerate(run_iter):
            batch_start = time.time()
            datatime += batch_start - batch_end

            output = model(batch)

            # from torchviz import make_dot, make_dot_from_trace
            # dot = make_dot(output.mean(), params=dict(model.named_parameters()))
            # pdb.set_trace()

            loss = float('NaN')
            if criterion:
                loss = criterion(output, getattr(batch, label_attr))

            if hasattr(batch, label_attr):
                scores = Runner._compute_scores(output, getattr(batch, label_attr))
            else:
                scores = [0] * 4

            cum_stats.update(float(loss), *scores)
            stats.update(float(loss), *scores)

            if return_predictions:
                for idx, id in enumerate(getattr(batch, id_attr)):
                    predictions.append((id, float(output[idx, 1].exp())))

            if (batch_idx + 1) % log_freq == 0:
                if progress_style == 'log':
                    Runner._print_stats(run_type, epoch + 1, batch_idx + 1, len(run_iter),
                                        stats, cum_stats)
                elif progress_style == 'tqdm-bar':
                    pbar.update()
                    Runner._set_pbar_status(pbar, stats, cum_stats)
                elif progress_style == 'bar':
                    pbar.update()
                stats = Statistics()

            if train:
                model.zero_grad()
                loss.backward()

                if not optimizer.params:
                    optimizer.set_parameters(model.named_parameters())
                optimizer.step()

            batch_end = time.time()
            runtime += batch_end - batch_start

        if progress_style == 'tqdm-bar':
            pbar.close()
        elif progress_style == 'bar':
            sys.stderr.flush()

        Runner._print_final_stats(epoch + 1, runtime, datatime, cum_stats)

        if return_predictions:
            return predictions
        else:
            return cum_stats.f1()

    @staticmethod
    def train(model,
              train_dataset,
              validation_dataset,
              best_save_path,
              epochs=30,
              criterion=None,
              optimizer=None,
              pos_neg_ratio=None,
              pos_weight=None,
              label_smoothing=0.05,
              save_every_prefix=None,
              save_every_freq=1,
              **kwargs):
        """run_train(model, train_dataset, validation_dataset, best_save_path,epochs=30, \
            criterion=None, optimizer=None, pos_neg_ratio=None, pos_weight=None, \
            label_smoothing=0.05, save_every_prefix=None, save_every_freq=None, \
            batch_size=32, device=None, progress_style='bar', log_freq=5, \
            sort_in_buckets=None)

        Train a :class:`deepmatcher.MatchingModel` using the specified training set.
        Refer to :meth:`deepmatcher.MatchingModel.run_train` for details on
        parameters.

        Returns:
            float: The best F1 score obtained by the model on the validation dataset.
        """

        model.initialize(train_dataset)

        model._register_train_buffer('optimizer_state', None)
        model._register_train_buffer('best_score', None)
        model._register_train_buffer('epoch', None)

        if criterion is None:
            if pos_weight is not None:
                assert pos_weight < 2
                warnings.warn('"pos_weight" parameter is deprecated and will be removed '
                              'in a later release, please use "pos_neg_ratio" instead',
                              DeprecationWarning)
                assert pos_neg_ratio is None
            else:
                if pos_neg_ratio is None:
                    pos_neg_ratio = 1
                else:
                    assert pos_neg_ratio > 0
                pos_weight = 2 * pos_neg_ratio / (1 + pos_neg_ratio)

            neg_weight = 2 - pos_weight

            criterion = SoftNLLLoss(label_smoothing,
                                    torch.Tensor([neg_weight, pos_weight]))

        optimizer = optimizer or Optimizer()
        if model.optimizer_state is not None:
            model.optimizer.base_optimizer.load_state_dict(model.optimizer_state)

        if model.epoch is None:
            epochs_range = range(epochs)
        else:
            epochs_range = range(model.epoch + 1, epochs)

        if model.best_score is None:
            model.best_score = -1
        optimizer.last_acc = model.best_score

        for epoch in epochs_range:
            model.epoch = epoch
            Runner._run(
                'TRAIN', model, train_dataset, criterion, optimizer, train=True, **kwargs)

            score = Runner._run('EVAL', model, validation_dataset, train=False, **kwargs)

            optimizer.update_learning_rate(score, epoch + 1)
            model.optimizer_state = optimizer.base_optimizer.state_dict()

            new_best_found = False
            if score > model.best_score:
                print('* Best F1:', score)
                model.best_score = score
                new_best_found = True

                if best_save_path and new_best_found:
                    print('Saving best model...')
                    model.save_state(best_save_path)
                    print('Done.')

            if save_every_prefix is not None and (epoch + 1) % save_every_freq == 0:
                print('Saving epoch model...')
                save_path = '{prefix}_ep{epoch}.pth'.format(
                    prefix=save_every_prefix, epoch=epoch + 1)
                model.save_state(save_path)
                print('Done.')
            print('---------------------\n')

        print('Loading best model...')
        model.load_state(best_save_path)
        print('Training done.')

        return model.best_score

    def eval(model, dataset, **kwargs):
        """eval(model, dataset, device=None, batch_size=32, progress_style='bar', log_freq=5,
            sort_in_buckets=None)

        Evaluate a :class:`deepmatcher.MatchingModel` on the specified dataset.
        Refer to :meth:`deepmatcher.MatchingModel.run_eval` for details on
        parameters.

        Returns:
            float: The F1 score obtained by the model on the dataset.
        """
        return Runner._run('EVAL', model, dataset, **kwargs)

    def predict(model, dataset, output_attributes=False, **kwargs):
        """predict(model, dataset, output_attributes=False, device=None, batch_size=32, \
            progress_style='bar', log_freq=5, sort_in_buckets=None)

        Use a :class:`deepmatcher.MatchingModel` to obtain predictions, i.e., match scores
        on the specified dataset.

        Returns:
            pandas.DataFrame: A pandas DataFrame containing tuple pair IDs (in the "id"
                column) and the corresponding match score predictions (in the
                "match_score" column). Will also include all attributes in the original
                CSV file of the dataset if `output_attributes` is True.
        """
        # Create a shallow copy of the model and reset embeddings to use vocab and
        # embeddings from new dataset.
        model = copy.deepcopy(model)
        model._reset_embeddings(dataset.vocabs)

        predictions = Runner._run(
            'PREDICT', model, dataset, return_predictions=True, **kwargs)
        pred_table = pd.DataFrame(predictions, columns=(dataset.id_field, 'match_score'))
        pred_table = pred_table.set_index(dataset.id_field)

        if output_attributes:
            raw_table = pd.read_csv(dataset.path).set_index(dataset.id_field)
            raw_table.index = raw_table.index.astype('str')
            pred_table = pred_table.join(raw_table)

        return pred_table
